from ..value_set import ValueSet


class AceInhibitorOrArbIngredient(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for medications of angiotensin-converting enzyme (ACE) inhibitor and angiotensin-receptor blocker (ARB) ingredients.

    **Data Element Scope:** This value set may use a model element related to Allergy/Intolerance.

    **Inclusion Criteria:** Includes concepts that represent an allergy or intolerance for prescribable angiotensin-converting enzyme (ACE) inhibitor and angiotensin-receptor blocker (ARB) ingredients only.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS135v10
    """

    VALUE_SET_NAME = 'ACE Inhibitor or ARB Ingredient'
    OID = '2.16.840.1.113883.3.526.3.1489'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    RXNORM = {
        '1998',  # captopril
        '3827',  # enalapril
        '18867',  # benazepril
        '29046',  # lisinopril
        '30131',  # moexipril
        '35208',  # quinapril
        '35296',  # ramipril
        '38454',  # trandolapril
        '50166',  # fosinopril
        '52175',  # losartan
        '54552',  # perindopril
        '69749',  # valsartan
        '73494',  # telmisartan
        '83515',  # eprosartan
        '83818',  # irbesartan
        '214354',  # candesartan
        '321064',  # olmesartan
    }


class BetaBlockerTherapyIngredient(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for medications containing ingredients for beta blocker therapy.

    **Data Element Scope:** This value set may use a model element related to Allergy/Intolerance.

    **Inclusion Criteria:** Includes concepts that represent an allergy or intolerance for prescribable beta blockers.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS144v10, CMS145v10
    """

    VALUE_SET_NAME = 'Beta Blocker Therapy Ingredient'
    OID = '2.16.840.1.113883.3.526.3.1493'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    RXNORM = {
        '149',  # acebutolol
        '1202',  # atenolol
        '1520',  # betaxolol
        '2116',  # carteolol
        '6185',  # labetalol
        '6918',  # metoprolol
        '7226',  # nadolol
        '8332',  # pindolol
        '8787',  # propranolol
        '10600',  # timolol
        '19484',  # bisoprolol
        '20352',  # carvedilol
    }


class EggSubstance(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for an allergy or intolerance to an egg substance.

    **Data Element Scope:** This value set may use a model element related to Substance.

    **Inclusion Criteria:** Includes concepts that represent an allergy or intolerance to an egg substance.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS147v11
    """

    VALUE_SET_NAME = 'Egg Substance'
    OID = '2.16.840.1.113883.3.526.3.1537'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    SNOMEDCT = {
        '102263004',  # Eggs (edible) (substance)
        '226881001',  # Dried egg (substance)
        '226885005',  # Raw egg (substance)
        '229955000',  # Dried egg white (substance)
        '256442007',  # Egg yolk (substance)
        '256443002',  # Egg white (substance)
        '286550009',  # Hen's egg (substance)
        '303300008',  # Egg protein (substance)
        '414074006',  # Egg product (substance)
    }


class InfluenzaVaccination(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for an influenza vaccination.

    **Data Element Scope:** This value set may use a model element related to Procedure.

    **Inclusion Criteria:** Includes concepts that represent an influenza vaccination.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS147v11
    """

    VALUE_SET_NAME = 'Influenza Vaccination'
    OID = '2.16.840.1.113883.3.526.3.402'
    DEFINITION_VERSION = '20170908'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CPT = {
        '90630',  # Influenza virus vaccine, quadrivalent (IIV4), split virus, preservative free, for intradermal use
        '90653',  # Influenza vaccine, inactivated (IIV), subunit, adjuvanted, for intramuscular use
        '90654',  # Influenza virus vaccine, trivalent (IIV3), split virus, preservative-free, for intradermal use
        '90655',  # Influenza virus vaccine, trivalent (IIV3), split virus, preservative free, 0.25 mL dosage, for intramuscular use
        '90656',  # Influenza virus vaccine, trivalent (IIV3), split virus, preservative free, 0.5 mL dosage, for intramuscular use
        '90657',  # Influenza virus vaccine, trivalent (IIV3), split virus, 0.25 mL dosage, for intramuscular use
        '90658',  # Influenza virus vaccine, trivalent (IIV3), split virus, 0.5 mL dosage, for intramuscular use
        '90661',  # Influenza virus vaccine, trivalent (ccIIV3), derived from cell cultures, subunit, preservative and antibiotic free, 0.5 mL dosage, for intramuscular use
        '90662',  # Influenza virus vaccine (IIV), split virus, preservative free, enhanced immunogenicity via increased antigen content, for intramuscular use
        '90666',  # Influenza virus vaccine (IIV), pandemic formulation, split virus, preservative free, for intramuscular use
        '90667',  # Influenza virus vaccine (IIV), pandemic formulation, split virus, adjuvanted, for intramuscular use
        '90668',  # Influenza virus vaccine (IIV), pandemic formulation, split virus, for intramuscular use
        '90673',  # Influenza virus vaccine, trivalent (RIV3), derived from recombinant DNA, hemagglutinin (HA) protein only, preservative and antibiotic free, for intramuscular use
        '90674',  # Influenza virus vaccine, quadrivalent (ccIIV4), derived from cell cultures, subunit, preservative and antibiotic free, 0.5 mL dosage, for intramuscular use
        '90682',  # Influenza virus vaccine, quadrivalent (RIV4), derived from recombinant DNA, hemagglutinin (HA) protein only, preservative and antibiotic free, for intramuscular use
        '90685',  # Influenza virus vaccine, quadrivalent (IIV4), split virus, preservative free, 0.25 mL dosage, for intramuscular use
        '90686',  # Influenza virus vaccine, quadrivalent (IIV4), split virus, preservative free, 0.5 mL dosage, for intramuscular use
        '90687',  # Influenza virus vaccine, quadrivalent (IIV4), split virus, 0.25 mL dosage, for intramuscular use
        '90688',  # Influenza virus vaccine, quadrivalent (IIV4), split virus, 0.5 mL dosage, for intramuscular use
        '90689',  # Influenza virus vaccine, quadrivalent (IIV4), inactivated, adjuvanted, preservative free, 0.25 mL dosage, for intramuscular use
        '90694',  # Influenza virus vaccine, quadrivalent (aIIV4), inactivated, adjuvanted, preservative free, 0.5 mL dosage, for intramuscular use
        '90756',  # Influenza virus vaccine, quadrivalent (ccIIV4), derived from cell cultures, subunit, antibiotic free, 0.5 mL dosage, for intramuscular use
    }
    HCPCSLEVELII = {
        'G0008',  # Administration of influenza virus vaccine
        'Q2034',  # Influenza virus vaccine, split virus, for intramuscular use (agriflu)
        'Q2035',  # Influenza virus vaccine, split virus, when administered to individuals 3 years of age and older, for intramuscular use (afluria)
        'Q2036',  # Influenza virus vaccine, split virus, when administered to individuals 3 years of age and older, for intramuscular use (flulaval)
        'Q2037',  # Influenza virus vaccine, split virus, when administered to individuals 3 years of age and older, for intramuscular use (fluvirin)
        'Q2038',  # Influenza virus vaccine, split virus, when administered to individuals 3 years of age and older, for intramuscular use (fluzone)
        'Q2039',  # Influenza virus vaccine, not otherwise specified
    }
    SNOMEDCT = {
        '86198006',  # Administration of vaccine product containing only Influenza virus antigen (procedure)
    }


class InfluenzaVaccine(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for immunizations with influenza vaccines.

    **Data Element Scope:** This value set may use a model element related to Immunization.

    **Inclusion Criteria:** Includes concepts that represent an influenza vaccine.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS147v11
    """

    VALUE_SET_NAME = 'Influenza Vaccine'
    OID = '2.16.840.1.113883.3.526.3.1254'
    DEFINITION_VERSION = '20170908'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CVX = {
        '88',  # influenza virus vaccine, unspecified formulation
        '135',  # influenza, high dose seasonal, preservative-free
        '140',  # Influenza, seasonal, injectable, preservative free
        '141',  # Influenza, seasonal, injectable
        '144',  # seasonal influenza, intradermal, preservative free
        '149',  # influenza, live, intranasal, quadrivalent
        '150',  # Influenza, injectable, quadrivalent, preservative free
        '155',  # Seasonal, trivalent, recombinant, injectable influenza vaccine, preservative free
        '158',  # influenza, injectable, quadrivalent, contains preservative
        '161',  # Influenza, injectable,quadrivalent, preservative free, pediatric
        '166',  # influenza, intradermal, quadrivalent, preservative free, injectable
        '168',  # Seasonal trivalent influenza vaccine, adjuvanted, preservative free
        '171',  # Influenza, injectable, Madin Darby Canine Kidney, preservative free, quadrivalent
        '185',  # Seasonal, quadrivalent, recombinant, injectable influenza vaccine, preservative free
        '186',  # Influenza, injectable, Madin Darby Canine Kidney,  quadrivalent with preservative
        '197',  # influenza, high-dose seasonal, quadrivalent, .7mL dose, preservative free
        '205',  # influenza, seasonal vaccine, quadrivalent, adjuvanted, .5mL dose, preservative free
    }


class StatinAllergen(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts of an allergy to statin medications.

    **Data Element Scope:** This value set may use a model element related to Allergy/Intolerance.

    **Inclusion Criteria:** Includes concepts that define an allergy to statin medications.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS347v5
    """

    VALUE_SET_NAME = 'Statin Allergen'
    OID = '2.16.840.1.113762.1.4.1110.42'
    DEFINITION_VERSION = '20210220'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    RXNORM = {
        '6472',  # lovastatin
        '36567',  # simvastatin
        '41127',  # fluvastatin
        '42463',  # pravastatin
        '83367',  # atorvastatin
        '301542',  # rosuvastatin
        '861634',  # pitavastatin
    }
    SNOMEDCT = {
        '372912004',  # Substance with 3-hydroxy-3-methylglutaryl-coenzyme A reductase inhibitor mechanism of action (substance)
    }




from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    EncounterToEvaluateBmi, Ethnicity, FollowUpForAboveNormalBmi, FollowUpForBelowNormalBmi,
    HospiceCareAmbulatory_1584, MedicationsForAboveNormalBmi, MedicationsForBelowNormalBmi,
    OncAdministrativeSex, PalliativeOrHospiceCare, Payer, PregnancyOrOtherRelatedDiagnoses, Race,
    ReferralsWhereWeightAssessmentMayOccur)


class ClinicalQualityMeasure69v9(ClinicalQualityMeasure):
    """
    Preventive Care and Screening: Body Mass Index (BMI) Screening and Follow-Up Plan

    Description: Percentage of patients aged 18 years and older with a BMI documented during the
    current encounter or within the previous twelve months AND who had a follow-up plan documented
    if most recent BMI was outside of normal parameters

    Definition: Normal BMI Parameters:  Age 18 years and older BMI >= 18.5 and < 25 kg/m2

    BMI- Body mass index (BMI) is a number calculated using the Quetelet index: weight divided by
    height squared (W/H2) and is commonly used to classify weight categories. BMI can be calculated
    using:

    Metric Units:  BMI = Weight (kg) / (Height (m) x Height (m))
    OR
    English Units:  BMI = Weight (lbs.) / (Height (in) x Height (in)) x 703

    Follow-Up Plan - Proposed outline of treatment to be conducted as a result of a BMI out of
    normal parameters. A follow-up plan may include, but is not limited to: documentation of
    education, referral (for example a Registered Dietitian Nutritionist (RDN), occupational
    therapist, physical therapist, primary care provider, exercise physiologist, mental health
    professional, or surgeon) for lifestyle/behavioral therapy, pharmacological interventions,
    dietary supplements, exercise counseling and/or nutrition counseling

    Rationale: BMI Above Normal Parameters

    Obesity is a chronic, multifactorial disease with complex psychological, environmental (social
    and cultural), genetic, physiologic, metabolic and behavioral causes and consequences. The
    prevalence of overweight and obese people is increasing worldwide at an alarming rate in both
    developing and developed countries. Environmental and behavioral changes brought about by
    economic development, modernization and urbanization have been linked to the rise in global
    obesity. The health consequences are becoming apparent (Fitch, 2013).

    Hales et al. (2017), report that the prevalence of obesity among adults and youth in the United
    States was 39.8% and 18.5% respectively, from 2015-2016. They note that obesity prevalence was
    higher among adults in the 40-59 age bracket than those in the 20-39 age bracket, for both men
    and women. Hales et al. (2017) also disaggregated the data according to ethnicity and noted
    that obesity prevalence was higher among non-Hispanic black and Hispanic adults and youth when
    compared with other races ethnicities. While obesity prevalence was lower among non-Hispanic
    Asian men and women, obesity prevalence among men, was comparable between non-Hispanic black
    and non-Hispanic white men. Obesity prevalence was higher among Hispanic men compared with non-
    Hispanic black men. While the prevalence among non-Hispanic black and Hispanic women was
    comparable, the prevalence for both groups was higher than that of non-Hispanic white women.
    Most notably, Hales et al. (2017), report that the prevalence of obesity in the United States
    remains higher than the Healthy People 2020 goals of 14.5% among youth and 30.5% among adults.

    More than a third of U.S. adults have a body mass index [BMI] >= 30 kg/m2; substantially at
    increased risk for diabetes and cardiovascular disease (CVD) (Flegal et al., 2012; Ogden et
    al., 2014). Behavioral weight management treatment has been identified as an effective first-
    line treatment for obesity with an average initial weight loss of eight to ten percent. This
    percentage weight loss is associated with a significant risk reduction for diabetes and CVD
    (Wadden, Butryn & Wilson, 2007). Despite the availability of effective interventions, two-
    thirds of obese U.S. patients were not offered or referred to weight management treatment
    during their primary care visit between 2005 and 2006, (Ma et al., 2009). In addition, the rate
    of weight management counseling in primary care significantly decreased by ten percent (40% to
    30%) between 1995-1996 and 2007-2008 (Kraschnewski et al., 2013). This suggests that the
    availability of evidence based clinical guidelines since 2008 obesity management in primary
    care remains suboptimal (Fitzpatrick & Stevens, 2017).

    BMI continues to be a common and reasonably reliable measurement to identify overweight and
    obese adults who may be at an increased risk for future morbidity. Although good quality
    evidence supports obtaining a BMI, it is important to recognize it is not a perfect
    measurement. BMI is not a direct measure of adiposity and as a consequence it can over or
    underestimate adiposity. BMI is a derived value that correlates well with total body fat and
    markers of secondary complications, e.g., hypertension and dyslipidemia (Barlow & the Expert
    Committee, 2007).

    In contrast with waist circumference, BMI and its associated disease and mortality risk appear
    to vary among ethnic subgroups. Female African American populations appear to have the lowest
    mortality risk at a BMI of 26.2-28.5 kg/m2 and 27.1-30.2 kg/m2 for women and men, respectively.
    In contrast, Asian populations may experience lowest mortality rates starting at a BMI of 23 to
    24 kg/m2. The correlation between BMI and diabetes risk also varies by ethnicity (LeBlanc et
    al., 2011, pp. 2-3).

    Screening for BMI and follow-up therefore is critical to closing this gap and contributes to
    quality goals of population health and cost reduction. However, due to concerns for other
    underlying conditions (such as bone health) or nutrition related deficiencies providers are
    cautioned to use clinical judgment and take these into account when considering weight
    management programs for overweight patients, especially the elderly (National Heart, Lung, and
    Blood Institute [NHLBI] Obesity Education Initiative, 1998, p. 91).

    It is important to enhance beneficiary access to all existing providers of Intensive Behavioral
    Therapy for obesity (IBT) which would result in decreased healthcare costs and lower obesity
    rates. Dietary counseling performed by a Registered Dietitian Nutritionist (RDN) is more
    effective than by a primary care clinician. IBT provided by RDNs for 6-12 months shows
    significant mean weight loss of up to 10% of body weight, maintained over one year’s time
    (Raynor & Champagne, 2016).

    BMI below Normal Parameters

    On the other end of the body weight spectrum is underweight (BMI < 18.5 kg/m2), which is
    equally detrimental to population health. When compared to normal weight individuals (BMI
    18.5-25 kg/m2), underweight individuals have significantly higher death rates with a Hazard
    Ratio of 2.27 and 95% confidence intervals (CI) = 1.78, 2.90 (Borrell & Lalitha, 2014).

    Poor nutrition or underlying health conditions can result in underweight (Fryar & Ogden, 2012).
    The National Health and Nutrition Examination Survey (NHANES) results from the 2007-2010
    indicate that women are more likely to be underweight than men. Therefore patients should be
    equally screened for underweight and followed up with nutritional counselling to reduce
    mortality and morbidity associated with underweight.

    Guidance: *  This eCQM is a patient-based measure. This measure is to be reported a minimum of
    once per reporting period for patients seen during the reporting period.
    *  There is no diagnosis associated with this measure.
    *  This measure may be reported by eligible professionals who perform the quality actions
    described in the measure based on the services provided at the time of the qualifying visit and
    the measure-specific denominator coding.

    BMI Measurement Guidance:
    *  Height and Weight - An eligible professional or their staff is required to measure both
    height and weight. Both height and weight must be measured within twelve months of the current
    encounter and may be obtained from separate encounters. Self-reported values cannot be used.
    *  The BMI may be documented in the medical record of the provider or in outside medical
    records obtained by the provider.
    *  If the most recent documented BMI is outside of normal parameters, then a follow-up plan is
    documented during the encounter or during the previous twelve months of the current encounter.
    *  If more than one BMI is reported during the measurement period, the most recent BMI will be
    used to determine if the performance has been met.
    *  Review the exclusions and exceptions criteria to determine those patients that BMI
    measurement may not be appropriate or necessary.

    Follow-Up Plan Guidance:

     * The documented follow-up plan must be based on the most recent documented BMI, outside of
    normal parameters, example: "Patient referred to nutrition counseling for BMI above or below
    normal parameters."

    (See Definitions for examples of follow-up plan treatments).

    Variation has been noted in studies exploring optimal BMI ranges for the elderly (see Donini et
    al., [2012]; Holme & Tonstad [2015]; Diehr et al. [2008]). Notably however, all these studies
    have arrived at ranges that differ from the standard range for ages 18 and older, which is
    >=18.5 and < 25 kg/m2. For instance, both Donini et al. (2012) and Holme and Tonstad (2015)
    reported findings that suggest that higher BMI (higher than the upper end of 25kg/m2) in the
    elderly may be beneficial. Similarly, worse outcomes have been associated with being
    underweight (at a threshold higher than 18.5 kg/m2) at age 65 (Diehr et al. 2008). Because of
    optimal BMI range variation recommendations from these studies, no specific optimal BMI range
    for the elderly is used. However, it may be appropriate to exempt certain patients from a
    follow-up plan by applying the exception criteria. See denominator exception section for
    examples.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms69v9
    """

    title = 'Preventive Care and Screening: Body Mass Index (BMI) Screening and Follow-Up Plan'

    identifiers = ['CMS69v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Barlow, S. E., & the Expert Committee. (2007). Expert committee recommendations regarding the prevention, assessment, and treatment of child and adolescent overweight and obesity: Summary report. Pediatrics, 120(Suppl. 4), S164-S192. doi:10.1542/peds.2007-2329C',
        'Borrell, L. N., & Samuel, L. (2014). Body mass index categories and mortality risk in U.S. adults: The effect of overweight and obesity on advancing death. American Journal of Public Health, 104(3), 512-519. doi:10.2105/AJPH.2013.301597',
        'Centers for Disease Control and Prevention (CDC). (2012). National Health and Nutrition Examination Survey (NHANES). Prevalence of underweight among adults aged 20 and over: United States, 1960–1962 Through 2011–2012. Retrieved from https://www.cdc.gov/nchs/data/hestat/underweight_adult_11_12/underweight_adult_11_12.htm',
        'Diehr, P., O’Meara, E. S., Fitzpatrick A., Newman, A. B., Kuller, L., Burke, G. (2008). Weight, mortality, years of healthy life, and active life expectancy in older adults. Journal of the American Geriatrics Society, 56(1), 76-83. doi:10.1111/j.1532-5415.01500.x',
        'Donini, L. M., Savina, C., Gennaro, E., De Felice, M. R., Rosano, A., Pandolfo, M. M., Del Balzo, V., …Chumlea, W. C. et al. (2012). A systematic review of the literature concerning the relationship between obesity and mortality in the elderly. The Journal of Nutrition, Health & Aging, 16(1), 89-98. doi:10.1007/s12603-011-0073-x',
        'Flegal, K. M., Carroll, M. D., Kit, B. K., & Ogden, C. L. (2012). Prevalence of obesity and trends in the distribution of body mass index among U. S. adults, 1999-2010. JAMA, 307(5), 491-497. doi.10.1001/jama.2012.39',
        'Fitch, A., Everling, L., Fox, C.,Goldberg, J., Heim, C., Johnson, K., …Webb, B. (2013, May). Prevention and management of obesity for adults. Bloomington, MN: Institute for Clinical Systems Improvement.',
        'Fitzpatrick, S. L., & Stevens, V. J. (2017, June 1). Adult obesity management in primary care, 2008-2013. Preventive Medicine, 99, 128-133. Retrieved from http://dx.doi.org/10.1016/j.ypmed.2017.02.020 doi:10.1016.j.ypmed.2017.01.020',
        'Fryar, C. D., & Ogden, C. L. (2012). Prevalence of underweight among adults aged 20 and over: United States, 1960-1962 through 2007-2010. Hyattsville, MD: NCHS, Division of Health and Nutrition Examination Surveys. Retrieved from http://www.cdc.gov/nchs/data/hestat/underweight_adult_07_10/underweight_adult_07_10.pdf',
        'Garvey, W. T., Mechanick, J. I., Brett, E. M., Garber, A. J., Hurley, D. L., Jastrebodd. A. M., …and Reviewers of the AACE/ACE Obesity Clinical Practice Guidelines. (2016). American Association of Clinical Endocrinologists and American College of Endocrinology comprehensive clinical practice guidelines for medical care of patients with obesity. Endocrine Practice, 22(Suppl. 3), 1-203. doi:10.4158/EP161365GL',
        'Hales, C. M., Carroll, M. D., Fryar, C. D., et al. (2017, October). Prevalence of obesity among adults and youth: United States, 2015-2016. NCHS Data Brief No. 288. Retrieved from https://www.cdc.gov/nchs/products/databriefs/db288.htm',
        'Holme, I., & Tonstad, S. (2015). Survival in elderly men in relation to midlife and current BMI. Age and Ageing, 44(3), 434-439.',
        'LeBlanc, E., O’Connor, E., Whitlock, E. P., et al. (2011). Screening for and management of obesity and overweight in adults (Evidence Report No. 89; AHRQ Publication No. 11-05159-EF-1). Rockville, MD: Agency for Healthcare Research and Quality.',
        'Kraschnewski, J. L., Sciamanna, C. N, Stuckey, H. L., et al. (2013, February). A silent response to the obesity epidemic: Decline in US physician weight counseling, 51 (2). Retrieved from http://prowellness.vmhost.psu.edu/wp-content/uploads/obesity_epidemic.pdf',
        'NHLBI Obesity Education Initiative. (1998). Clinical guidelines on the identification, evaluation, and treatment of overweight and obesity in adults (Report No. 98-4083). Bethesda, MD: NHLBI.',
        'Ogden, C.L., Carroll, M.D., Fryar, C.D., Flegal, K.M. (2015). Prevalence of obesity among adults and youth: United States, 2011–2014. NCHS data brief, no 219. Hyattsville, MD: National Center for Health Statistics. Retrieved from https://www.cdc.gov/nchs/data/databriefs/db219.pdf',
        'Raynor, H. A., & Champagne, C. M. (2016). Position of the Academy of Nutrition and Dietetics: Interventions for the treatment of overweight and obesity in adults. Journal of the Academy of Nutrition and Dietetics, 116(1), 129-147. doi:10.1016/jand.2015.10.031',
        'U.S. Preventive Services Task Force. (2018). Behavioral weight loss interventions to prevent obesity-related morbidity and mortality in adults: U.S. Preventive Services Task Force recommendation statement. JAMA, 320(11), 1163–1171. doi:10.1001/jama.2018.13022',
        'Wadden, T. A, Butryn, M. L., Wilson, C. (2007). Lifestyle modification for the management of obesity. Gastroenterology, 132 (6), 2226-2238. doi: 10.1053/j.gastro.2007.03.051',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients aged 18 and older on the date of the encounter with at
        least one eligible encounter during the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Patients who are pregnant
        Patients receiving palliative or hospice care

        Exceptions: Patients with a documented medical reason for not documenting BMI or for not
        documenting a follow-up plan for a BMI outside normal parameters (e.g., elderly patients
        (65 or older) for whom weight reduction/weight gain would complicate other underlying
        health conditions such as illness or physical disability, mental illness, dementia,
        confusion, or nutritional deficiency such as vitamin/mineral deficiency; patients in an
        urgent or emergent medical situation where time is of the essence and to delay treatment
        would jeopardize the patient’s health status)

        Patients who refuse measurement of height and/or weight
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients with a documented BMI during the encounter or during the previous
        twelve months, AND when the BMI is outside of normal parameters, a follow-up plan is
        documented during the encounter or during the previous twelve months of the current
        encounter

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: All adults should be screened annually using a BMI measurement.
        BMI measurements >= 25 kg/m2 should be used to initiate further evaluation of overweight or
        obesity after taking into account age, gender, ethnicity, fluid status, and muscularity;
        therefore, clinical evaluation and judgment must be used when BMI is employed as the
        anthropometric indicator of excess adiposity, particularly in athletes and those with
        sarcopenia (Garvey et al., 2016 AACE/ACE Guidelines, 2016, pp. 12-13) (Grade A).

        Overweight and Underweight Categories:
        Underweight < 18.5; Normal weight 18.5-24.9; Overweight 25-29.9; Obese class I 30-34.9;
        Obese class II 35-39.9; Obese class III >= 40 (Garvey et al., 2016 AACE/ACE Guidelines,
        2016, p. 15).

        BMI cutoff point value of >= 23 kg/m2 should be used in the screening and confirmation of
        excess adiposity in Asian adults (Garvey et al., 2016 AACE/ACE Guidelines, 2016, p. 13)
        (Grade B).

        Lifestyle/Behavioral Therapy for Overweight and Obesity should include behavioral
        interventions that enhance adherence to prescriptions for a reduced-calorie meal plan and
        increased physical activity (behavioral interventions can include: self-monitoring of
        weight, food intake, and physical activity; clear and reasonable goal-setting; education
        pertaining to obesity, nutrition, and physical activity; face-to-face and group meetings;
        stimulus control; systematic approaches for problem solving; stress reduction; cognitive
        restructuring [i.e., cognitive behavioral therapy], motivational interviewing; behavioral
        contracting; psychological counseling; and mobilization of social support structures)
        (Garvey et al., 2016 AACE/ACE Guidelines, 2016, p. 22) (Grade A).

        Behavioral lifestyle intervention should be tailored to a patient's ethnic, cultural,
        socioeconomic, and educational background (Garvey et al., 2016 AACE/ACE Guidelines, 2016,
        p. 22) (Grade B).

        The USPSTF recommends that clinicians offer or refer adults with a body mass index (BMI) of
        30 kg/m2 or higher  to intensive, multicomponent behavioral interventions  (USPSTF, 2018)
        (Grade B).

        Interventions:
        - Effective intensive behavioral interventions were designed to help participants achieve
        or maintain a >= 5% weight loss through a combination of dietary changes and increased
        physical activity
        - Most interventions lasted for 1 to 2 years, and the majority had >= 12 sessions in the
        first year
        - Most behavioral interventions focused on problem solving to identify barriers, self-
        monitoring of weight, peer support, and relapse prevention
        - Interventions also provided tools to support weight loss or weight loss maintenance
        (e.g., pedometers, food scales, or exercise videos) (USPSTF, 2018)

        Nutritional safety for the elderly should be considered when recommending weight reduction.
        "A clinical decision to forego obesity treatment in older adults should be guided by an
        evaluation of the potential benefits of weight reduction for day-to-day functioning and
        reduction of the risk of future cardiovascular events, as well as the patient's motivation
        for weight reduction. Care must be taken to ensure that any weight reduction program
        minimizes the likelihood of adverse effects on bone health or other aspects of nutritional
        status" (NHLBI Obesity Education Initiative, 1998, p. 91) (Evidence Category D). In
        addition, weight reduction prescriptions in older persons should be accompanied by proper
        nutritional counseling and regular body weight monitoring (NHLBI Obesity Education
        Initiative, 1998, p. 91).

        The possibility that a standard approach to weight loss will work differently in diverse
        patient populations must be considered when setting expectations about treatment outcomes
        (NHLBI Obesity Education Initiative, 1998, p. 97) (Evidence Category B).
        """
        pass



from ..value_set import ValueSet


class EggSubstance(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent an allergy or intolerance to an egg substance.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) category related to Allergy/Intolerance.

    **Inclusion Criteria:** Includes only relevant concepts associated with an allergy or intolerance to an egg substance.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1537'
    VALUE_SET_NAME = 'Egg Substance'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '102263004',
        '226881001',
        '226885005',
        '229955000',
        '256442007',
        '256443002',
        '286550009',
        '303300008',
        '414074006'
    }


from unittest import TestCase

from canvas_workflow_kit.canvas_code_set import CanvasCodeSet


class TestCanvasCodeSet(TestCase):

    def test_codes(self):
        expected = {
            'CANVAS0001': {
                'code': 'CANVAS0001',
                'description': 'Questionnaire used to determine if the patient has sufficient risk for COVID-19 to require a follow-up.',
                'short_description': 'COVID-19 Follow-up qualification questionnaire',
            },
            'CANVAS0002': {
                'code': 'CANVAS0002',
                'description': 'Question used to determine if the patient has exhibited symptoms commonly associated with COVID-19.',
                'short_description': 'COVID-19 Symptoms Question',
            },
            'CANVAS0003': {
                'code': 'CANVAS0003',
                'description': 'Question used to determine if the patient has traveled within the past 14 days.',
                'short_description': 'COVID-19 Recent Travel Question',
            },
            'CANVAS0004': {
                'code': 'CANVAS0004',
                'description': 'Question used to determine if the patient has been exposed to someone with a confirmed case of COVID-19.',
                'short_description': 'COVID-19 Exposure Question',
            },
            'CANVAS0005': {
                'code': 'CANVAS0005',
                'description': 'Question used to override COVID-19 risk scoring and force a follow-up.',
                'short_description': 'COVID-19 Forced Follow-up Question',
            },
            'CANVAS0006': {
                'code': 'CANVAS0006',
                'description': 'Questionnaire used to check in on patients with higher risk for COVID-19',
                'short_description': 'COVID-19 High Risk Check-in Questionnaire',
            },
            'CANVAS0007': {
                'code': 'CANVAS0007',
                'description': "Question used to gauge patient's knowledge of the Coronavirus situation.",
                'short_description': 'COVID-19 Knowledge Check Question',
            },
            'CANVAS0008': {
                'code': 'CANVAS0008',
                'description': 'Question used to check which COVID-19 prevention steps the patient has started using.',
                'short_description': 'COVID-19 Preventive Steps Employed Question',
            },
            'CANVAS0009': {
                'code': 'CANVAS0009',
                'description': 'Question used to ask if the patient has any questions about COVID-19 prevention.',
                'short_description': 'Question About COVID-19 Prevention Questions',
            },
            'CANVAS0010': {
                'code': 'CANVAS0010',
                'description': 'Question used to determine patient worry about COVID-19 impacts.',
                'short_description': 'COVID-19 Impact Worry Question',
            },
            'CANVAS0011': {
                'code': 'CANVAS0011',
                'description': 'Question used to identify patient social isolation mitigation strategy.',
                'short_description': 'COVID-19 Social Isolation Mitigation Strategy Question',
            },
            'CANVAS0012': {
                'code': 'CANVAS0012',
                'description': 'Question asking what help the patient would like from the clinic.',
                'short_description': 'Question asking what help the patient would like from the clinic.',
            },
            'CANVAS0013': {
                'code': 'CANVAS0013',
                'description': 'Question used to note the internal assessment of the COVID-19 outreach call.',
                'short_description': 'COVID-19 Outreach Call Internal Assessment',
            },
        }
        assert expected == CanvasCodeSet.CODES, 'We cannot MODIFY any existing code, adding is ok if you add new codes here'

    def test_code(self):
        tests = [
            ('CANVAS0001', 'CANVAS0001'),
            ('CANVAS0002', 'CANVAS0002'),
            ('CANVAS0003', 'CANVAS0003'),
            ('CANVAS0004', 'CANVAS0004'),
            ('CANVAS0005', 'CANVAS0005'),
            ('CANVAS0006', 'CANVAS0006'),
            ('CANVAS0007', 'CANVAS0007'),
            ('CANVAS0008', 'CANVAS0008'),
            ('CANVAS0009', 'CANVAS0009'),
            ('CANVAS0010', 'CANVAS0010'),
            ('CANVAS0011', 'CANVAS0011'),
            ('CANVAS0012', 'CANVAS0012'),
            ('CANVAS0013', 'CANVAS0013'),
            ('CANVAS0016', ''),
            ('NOPE', ''),
        ]
        for key, expected in tests:
            result = CanvasCodeSet.code(key)
            self.assertEqual(expected, result)



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AllergyToEggs, AllergyToInfluenzaVaccine, AnnualWellnessVisit,
    CareServicesInLongTermResidentialFacility, DischargeServicesNursingFacility, EggSubstance,
    EncounterInfluenza, Ethnicity, HomeHealthcareServices, InfluenzaVaccination,
    InfluenzaVaccine_1254, IntoleranceToInfluenzaVaccine, NursingFacilityVisit, OfficeVisit,
    OncAdministrativeSex, OutpatientConsultation, PatientProviderInteraction, Payer,
    PreventiveCareEstablishedOfficeVisit0To17, PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesGroupCounseling, PreventiveCareServicesIndividualCounseling,
    PreventiveCareServicesInitialOfficeVisit0To17, PreventiveCareServicesInitialOfficeVisit18AndUp,
    PreventiveCareServicesOther, PreviousReceiptOfInfluenzaVaccine, Race)


class ClinicalQualityMeasure147v10(ClinicalQualityMeasure):
    """
    Preventive Care and Screening: Influenza Immunization

    Description: Percentage of patients aged 6 months and older seen for a visit between October 1
    and March 31 who received an influenza immunization OR who reported previous receipt of an
    influenza immunization

    Definition: Previous Receipt - receipt of the current season's influenza immunization from
    another provider OR from same provider prior to the visit to which the measure is applied
    (typically, prior vaccination would include influenza vaccine given since August 1st)

    Rationale: Influenza vaccination is the most effective protection against influenza virus
    infection (Centers for Disease Control and Prevention [CDC], 2018). Influenza may lead to
    serious complications including hospitalization or death (CDC, 2018). Influenza vaccine is
    recommended for all persons aged >=6 months who do not have contraindications to vaccination.
    However, data indicate that less than half of all eligible individuals receive an influenza
    vaccination (CDC, 2015). This measure promotes annual influenza vaccination for all persons
    aged >= 6 months.

    Guidance: The timeframe for the visit during the "Encounter, Performed": "Encounter-Influenza"
    or "Procedure, Performed": "Peritoneal Dialysis" or "Procedure, Performed": "Hemodialysis" in
    the Population Criteria-Denominator, refers to the influenza season defined by the measure:
    October through March (October 1 for the year prior to the start of the reporting period
    through March 31 during the reporting period). The "Encounter-Influenza" Grouping OID detailed
    in the data criteria section below is comprised of several individual OIDs of different
    encounter types. The individual OIDs are included in the value set and should be reviewed to
    determine that an applicable visit occurred during the timeframe for "Encounter, Performed":
    "Encounter-Influenza" as specified in the denominator.

    To enable reporting of this measure at the close of the reporting period, this measure will
    only assess the influenza season that ends in March of the reporting period. The subsequent
    influenza season (ending March of the following year) will be measured and reported in the
    following year.

    Due to the changing stance of the CDC/ACIP recommendations regarding the live attenuated
    influenza vaccine (LAIV) for a particular flu season, this measure will not include the
    administration of this specific formulation of the flu vaccination. Given the variance of the
    timeframes for the annual update cycles, program implementation, and publication of revised
    recommendations from the CDC/ACIP, it has been determined that the coding for this measure will
    specifically exclude this formulation, so as not to inappropriately include this form of the
    vaccine for flu seasons when CDC/ACIP explicitly advise against it. However, it is recommended
    that all eligible professionals or eligible clinicians to review the guidelines for each flu
    season to determine appropriateness of the LAIV and other formulations of the flu vaccine.
    Should the LAIV be recommended for administration for a particular flu season, eligible
    professional or clinician may consider one of the following options: 1) satisfy the numerator
    by reporting either previous receipt or using the CVX 88 for unspecified formulation, 2) report
    a denominator exception, either as a patient reason (e.g., for patient preference) or a system
    reason (e.g., the institution only carries LAIV).

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms147v10
    """

    title = 'Preventive Care and Screening: Influenza Immunization'

    identifiers = ['CMS147v10']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Centers for Disease Control and Prevention. (2018, October 26). About Flu. Retrieved from http://www.cdc.gov/flu/about/disease/index.htm',
        'Centers for Disease Control and Prevention. (2015, September 17). Flu vaccination coverage: United States, 2014-15 influenza season. Retrieved from http://www.cdc.gov/flu/fluvaxview/coverage-1415estimates.htm',
        'Grohskopf, L.A., Alyanak, E., Broder, K. R., Walter, E.B., Fry, A. M., Jernigan, D. B. (2019). Prevention and control of seasonal influenza with vaccines: Recommendations of the Advisory Committee on Immunization Practices—United States, 2019-20 Influenza Season. MMWR Recommendations and Reports; 38(No. RR-3), 1-21. doi: http://dx.doi.org/10.15585/mmwr.rr6803a1',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients aged 6 months and older seen for a visit during the
        measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population and seen for a visit between October 1 and March 31

        Exclusions: None

        Exceptions: Documentation of medical reason(s) for not receiving influenza immunization
        (e.g., patient allergy, other medical reasons).

        Documentation of patient reason(s) for not receiving influenza immunization (e.g., patient
        declined, other patient reasons).

        Documentation of system reason(s) for not receiving influenza immunization (e.g., vaccine
        not available, other system reasons).
        """
        pass

    def in_numerator(self):
        """
        Numerator: Patients who received an influenza immunization OR who reported previous receipt
        of an influenza immunization

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: Routine annual influenza vaccination is recommended for all
        persons aged >= 6 months who do not have contraindications. Optimally, vaccination should
        occur before onset of influenza activity in the community. Although vaccination by the end
        of October is recommended, vaccine administered in December or later, even if influenza
        activity has already begun, is likely to be beneficial in the majority of influenza seasons
        (CDC/Advisory Committee on Immunization Practices [ACIP], 2019).
        """
        pass



# type: ignore
import arrow

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import InstructionRecommendation
from canvas_workflow_kit.value_set.specials import Hcc005v1AnnualWellnessVisit


class Hcc005v1(ClinicalQualityMeasure):

    class Meta:
        title = 'Annual Wellness Visit'
        version = '2019-11-04v1'
        description = 'Patient 65 or older due  for Annual Wellness Visit.'
        information = 'https://canvas-medical.zendesk.com/hc/en-us/articles/360059083973-Annual-Wellness-Visit-HCC005v1'  # noqa: E501

        identifiers = ['HCC005v1']

        types = ['HCC']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]
        authors = [
            'Canvas Medical Team',
        ]

        references = [
            'Canvas Medical HCC, https://canvas-medical.zendesk.com/hc/en-us/articles/360059083973-Annual-Wellness-Visit-HCC005v1'  # noqa: E501
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_BILLING_LINE_ITEM,
            ClinicalQualityMeasure.CHANGE_PATIENT,
        ]

    MINIMUM_AGE = 65

    _last_visit = None

    def in_initial_population(self) -> bool:
        """
        Initial population: Patients 65+ years of age
        """
        return bool(self.patient.age_at(self.timeframe.end) > self.MINIMUM_AGE)

    def in_denominator(self) -> bool:
        """
        Patients in the initial population
        """
        return self.in_initial_population()

    def in_numerator(self) -> bool:
        """
        Patients without visit including cpt G0438 or G0439 in >1 year
        """
        self._last_visit = (
            self.patient.billing_line_items.find(Hcc005v1AnnualWellnessVisit).within(
                self.timeframe).last())
        return not bool(self._last_visit)

    def recent_visit_context(self) -> str:
        record = self.patient.billing_line_items.find(Hcc005v1AnnualWellnessVisit).last()
        if record:
            last_date = arrow.get(record['created'])
            return f'Last Annual Wellness Visit was {self.display_date(last_date)}.'
        else:
            return 'There are no Annual Wellness Visits on record.'

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        if self.in_denominator():
            if self.in_numerator():
                result.due_in = -1
                result.status = STATUS_DUE
                result.add_narrative(
                    f'{self.patient.first_name} is due for a Annual Wellness Visit.')
                result.add_narrative(self.recent_visit_context())
                result.add_recommendation(
                    InstructionRecommendation(
                        key='HCC005v1_RECOMMEND_WELLNESS_VISIT',
                        rank=1,
                        button='Plan',
                        patient=self.patient,
                        instruction=Hcc005v1AnnualWellnessVisit,
                        title='Schedule for Annual Wellness Visit'))
            else:
                visit_date = arrow.get(self._last_visit['created'])
                result.due_in = (visit_date.shift(days=self.timeframe.duration) - self.now).days
                result.status = STATUS_SATISFIED
                result.add_narrative('{patient} had a visit {date}.'.format(
                    patient=self.patient.first_name, date=self.display_date(visit_date)))
        else:
            result.due_in = (
                self.patient.birthday.shift(years=self.MINIMUM_AGE) - self.timeframe.end).days

        return result



from typing import List

import arrow

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import ClinicalQualityMeasure
from canvas_workflow_kit.recommendation import InterviewRecommendation, InstructionRecommendation, Recommendation
from canvas_workflow_kit.value_set.v2018 import (
    AdditionalEvaluationForDepressionAdolescent, AdditionalEvaluationForDepressionAdult,
    AdolescentDepressionScreening, AdultDepressionScreening, BipolarDiagnosis, DepressionDiagnosis,
    DepressionMedicationsAdolescent, DepressionMedicationsAdult, DepressionScreeningEncounterCodes,
    Ethnicity, FollowUpForDepressionAdolescent, FollowUpForDepressionAdult,
    NegativeDepressionScreening, OncAdministrativeSex, Payer, PositiveDepressionScreening, Race,
    ReferralForDepressionAdolescent, ReferralForDepressionAdult,
    SuicideRiskAssessment_559)  # flake8: noqa


class ClinicalQualityMeasure2v7(ClinicalQualityMeasure):
    """
    Preventive Care and Screening: Screening for Depression and Follow-Up Plan

    Description: Percentage of patients aged 12 years and older screened for depression on the date
    of the encounter using an age appropriate standardized depression screening tool AND if
    positive, a follow-up plan is documented on the date of the positive screen

    Definition: Screening:
    Completion of a clinical or diagnostic tool used to identify people at risk of developing or
    having a certain disease or condition, even in the absence of symptoms.
    Standardized Depression Screening Tool - A normalized and validated depression screening tool
    developed  for the patient population in which it is being utilized

    Examples of depression screening tools include but are not limited to:
    *  Adolescent Screening Tools (12-17 years)
       *  Patient Health Questionnaire for Adolescents (PHQ-A)
       *  Beck Depression Inventory-Primary Care Version (BDI-PC)
       *  Mood Feeling Questionnaire(MFQ)
       *  Center for Epidemiologic Studies Depression Scale (CES-D)
       *  Patient Health Questionnaire (PHQ-9)
       *  Pediatric Symptom Checklist (PSC-17)
       *  PRIME MD-PHQ2
    *  Adult Screening Tools (18 years and older)
       *  Patient Health Questionnaire (PHQ9)
       *  Beck Depression Inventory (BDI or BDI-II)
       *  Center for Epidemiologic Studies Depression Scale (CES-D)
       *  Depression Scale (DEPS)
       *  Duke Anxiety-Depression Scale (DADS)
       *  Geriatric Depression Scale (GDS)
       *  Cornell Scale for Depression in Dementia (CSDD)
       *  PRIME MD-PHQ2
       * Hamilton Rating Scale for Depression (HAM-D)
       * Quick Inventory of Depressive Symptomatology Self-Report (QID-SR)
    * Perinatal Screening Tools
       *  Edinburgh Postnatal Depression Scale
       *  Postpartum Depression Screening Scale
       *  Patient Health Questionnaire 9 (PHQ-9)
       *  Beck Depression Inventory
       *  Beck Depression Inventory-II
       *  Center for Epidemiologic Studies Depression Scale
       *  Zung Self-rating Depression Scale

    Follow-Up Plan:
    Documented follow-up for a positive depression screening must include one or more of the
    following:
     *  Additional evaluation for depression
     *  Suicide Risk Assessment
     *  Referral to a practitioner who is qualified to diagnose and treat depression
     *  Pharmacological interventions
     *  Other interventions or follow-up for the diagnosis or treatment of depression

    Rationale: 2014 U.S. survey data indicate that 2.8 million (11.4 percent) adolescents aged 12
    to 17 had a major depressive episode (MDE) in the past year and that 15.7 million (6.6 percent)
    adults aged 18 or older had at least one MDE in the past year, with 10.2 million adults (4.3
    percent) having one MDE with severe impairment in the past year (Center for Behavioral Health
    Statistics and Quality, 2015). The World Health Organization (WHO), as cited by Pratt & Brody
    (2008), found that major depression was the leading cause of disability worldwide. Data
    indicate that approximately 80% of people diagnosed with depression report some level of
    difficulty in functioning because of their depressive symptoms. For example, 35% of males and
    22% of females with depression reported that their depressive symptoms make it extremely
    difficult for them to work, get things done at home, or get along with other people.
    Additionally, more than one-half of all persons with mild depressive symptoms also reported
    some difficulty in daily functioning attributable to their depressive symptoms (Pratt & Brody,
    2008). In young adulthood, major depressive disorder (MDD) has been found to be associated with
    early pregnancy, decreased school performance, and impaired work, social, and family
    functioning (Williams et al., 2009, p. e716). In the perinatal period, depression and other
    mood disorders, such as bipolar disorder and anxiety disorders, can have devastating effects on
    women, infants, and families. Maternal suicide rates rise over hemorrhage and hypertensive
    disorders as a cause of maternal mortality (American College of Obstetricians and
    Gynecologists, 2015).

    Negative outcomes associated with depression make it crucial to screen in order to identify and
    treat depression in its early stages. While Primary Care Providers (PCPs) serve as the first
    line of defense in the detection of depression, studies show that PCPs fail to recognize up to
    50% of depressed patients (Borner, 2010, p. 948). 'Coyle et al. (2003), suggested that the
    picture is more grim for adolescents, and that more than 70% of children and adolescents
    suffering from serious mood disorders go unrecognized or inadequately treated' (Borner, 2010,
    p. 948). 'In nationally representative U.S. surveys, about 8% of adolescents reported having
    major depression in the past year. Only 36% to 44% of children and adolescents with depression
    receive treatment, suggesting that the majority of depressed youth are undiagnosed and
    untreated' (Sui, A. and USPSTF, 2016). Evidence supports that screening for depression in
    pregnant and postpartum women is of moderate net benefit and treatment options for positive
    depression screening should be available for patients twelve and older including pregnant and
    postpartum women.

    If preventing negative patient outcomes is not enough, the substantial economic burden of
    depression for individuals and society alike makes a case for screening for depression on a
    regular basis. Depression imposes economic burden through direct and indirect costs. 'In the
    United States, an estimated $22.8 billion was spent on depression treatment in 2009, and lost
    productivity cost an additional estimated $23 billion in 2011' (Sui, A. and USPSTF, 2016).

    This measure seeks to align with clinical guideline recommendations as well as the Healthy
    People 2020 recommendation for routine screening for mental health problems as a part of
    primary care for both children and adults (U.S. Department of Health and Human Services, 2014)
    and  makes an important contribution to the quality domain of community and population health.

    Guidance: A depression screen is completed on the date of the encounter using an age
    appropriate standardized depression screening tool AND if positive, either additional
    evaluation for depression, suicide risk assessment, referral to a practitioner who is qualified
    to diagnose and treat depression, pharmacological interventions, or other interventions or
    follow-up for the diagnosis or treatment of depression is documented on the date of the
    positive screen.

    Screening Tools:
     *  The name of the age appropriate standardized depression screening tool utilized must be
    documented in the medical record
     *  The depression screening must be reviewed and addressed in the office of the provider,
    filing the code, on the date of the encounter
     *  The screening should occur during a qualified encounter
     *  Standardized Depression Screening Tools should be normalized and validated for the age
    appropriate patient population in which they are used and must be documented in the medical
    record

    Follow-Up Plan:

    * The follow-up plan must be related to a positive depression screening, example: 'Patient
    referred for psychiatric evaluation due to positive depression screening.'

    * Pharmacologic treatment for depression is often indicated during pregnancy and/or lactation.
    Review and discussion of the risks of untreated versus treated depression is advised.
    Consideration of each patient's prior disease and treatment history, along with the risk
    profiles for individual pharmacologic agents, is important when selecting pharmacologic therapy
    with the greatest likelihood of treatment effect.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms2v7
    """

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'Quality Insights',
    ]

    references = [
        'American College of Obstetricians and Gynecologists (2015). Screening for perinatal depression. Committee Opinion No. 630. Obstet Gynecol 2015;125:1268-71. Retrieved from http://www.acog.org/Resources-And-Publications/Committee-Opinions/Committee-on-Obstetric-Practice/Screening-for-Perinatal-Depression',
        'Borner I, Braunstein JW, St. Victor, R, Pollack J (2010). Evaluation of a 2-question screening tool for detecting depression in adolescents in Primary Care. Clinical Pediatrics, 49, 947-995. doi: 10.1177/0009922810370203',
        'Center for Behavioral Health Statistics and Quality. (2015). Behavioral health trends in the United States:Results from the 2014 National Survey on Drug Use and Health (HHS Publication No. SMA 15-4927, NSDUH Series H-50). Retrieved from http://www.samhsa.gov/data/sites/default/files/NSDUH-FRR1-2014/NSDUH-FRR1-2014.pdf',
        'Coyle J T, Pine D.S, Charney D S, Lewis L, Nemeroff C B, Carlson G A, Joshi P T (2003). Depression and bipolar support alliance consensus development panel. Depression and bipolar support alliance consensus statement on the unmet needs in diagnosis and treatment of mood disorders in children and adolescents. Journal of the American Academy of Child and Adolescent Psychiatry, 42, 1494-1503.',
        'Pratt L.A, Brody DJ.(2008). Depression in the United States household population, 2005-2006. U.S. Department of Health and Human Services, Centers for Disease Control and Prevention National Center for Health Statistics. NCHS Data Brief No.7, 1-8.',
        'Siu AL, on behalf of the U.S. Preventive Services Task Force. Screening for Depression in Children and Adolescents: U.S. Preventive Services Task Force Recommendation Statement. Ann Intern Med. 2016;164:360-366. http://annals.org/article.aspx?articleid=2490528',
        'Siu AL, and the US Preventive Services Task Force (USPSTF). Screening for Depression in Adults: US Preventive Services Task Force Recommendation Statement. JAMA. 2016;315(4):380-387. doi:10.1001/jama.2015.18392. http://jama.jamanetwork.com/article.aspx?articleid=2484345',
        'Steinman LE, Frederick JT, Prohaska T, Satariano WA, Dornberg-Lee S, Fisher R, ...Snowden M (2007). Recommendations for treating depression in community-based older adults. American Journal of Preventive Medicine, 33(3), 175-81. Retrieved from: www.ajpm-online.net/article/S0749-3797%2807%2900330-3/abstract',
        'Trangle M, Gursky J, Haight R, Hardwig J, Hinnenkamp T, Kessler D, Mack N, Myszkowski M. Institute for Clinical Systems Improvement. Adult Depression in Primary Care. Updated March 2016. https://www.icsi.org/_asset/fnhdm3/Depr.pdf',
        'U.S. Department of Health and Human Services (2014). Healthy People 2020. Washington, DC: U.S. Department of Health and Human Services. Retrieved from: http://www.healthypeople.gov/2020/topicsobjectives2020/objectiveslist.aspx?topicId=28',
        'Wilkinson J, Bass C, Diem S, Gravley A, Harvey L, Maciosek M, McKeon K, Milteer L, Owens J, Rothe P, Snellman L, Solberg L, Vincent P. Institute for Clinical Systems Improvement. Preventive Services for Children and Adolescents. Updated September 2013. https://www.icsi.org/_asset/x1mnv1/PrevServKids.pdf',
        'Zalsman G, Brent DA & Weersing VR (2006). Depressive disorders in childhood and adolescence: an overview: epidemiology, clinical manifestation and risk factors. Child Adolesc Psychiatr Clin N Am. 2006;15:827-841',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients aged 12 years and older before the beginning of the
        measurement period with at least one eligible encounter during the measurement period
        """
        age = self.patient.age_at(self.timeframe.end)
        if age >= 12:
            return True
        return False

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Patients with an active diagnosis for depression or a diagnosis of bipolar
        disorder

        Exceptions: Patient Reason(s)
        Patient refuses to participate
        OR
        Medical Reason(s)
        Patient is in an urgent or emergent situation where time is of the essence and to delay
        treatment would jeopardize the patient's health status
        OR
        Situations where the patient's functional capacity or motivation to improve may impact the
        accuracy of results of standardized depression assessment tools.  For example: certain
        court appointed cases or cases of delirium
        """
        if not self.in_initial_population():
            return False

        if self.patient.conditions.find(BipolarDiagnosis | DepressionDiagnosis):
            return False

        # FIX: how do we find refusal to participate?
        # FIX: how do we look for the medical reasons listed above?

        return True

    def in_numerator(self):
        """
        Numerator: Patients screened for depression on the date of the encounter  using an age
        appropriate standardized tool AND if positive, a follow-up plan is documented on the date
        of the positive screen

        Exclusions: Not Applicable
        """
        screening = self.patient.interviews.find(AdolescentDepressionScreening |
                                                 AdultDepressionScreening).within(
                                                     self.timeframe).last()
        if not screening:
            return False

        # FIX: what's the right way to see if the screening was
        # positive, given that there are many possible screening
        # interviews?  This works but depends on this wording being
        # used for all the narratives.
        if 'positive' in screening['results'][0]['narrative'].lower():
            if not self.patient.instructions.find(FollowUpForDepressionAdolescent |
                                                  FollowUpForDepressionAdult).within(
                                                      self.timeframe):
                return False

        return True

    def compute_results(self):
        """
        Clinical recommendation: Adolescent Recommendation (12-18 years):

        'The USPSTF recommends screening for MDD in adolescents aged 12 to 18 years. Screening
        should be implemented with adequate systems in place to ensure accurate diagnosis,
        effective treatment, and appropriate follow-up (B recommendation)' (Sui, A. and USPSTF,
        2016, p. 360).

        'Clinicians and health care systems should try to consistently screen adolescents, ages
        12-18,  for major depressive disorder, but only when systems are in place to ensure
        accurate diagnosis, careful selection of treatment, and close follow-up' (ICSI, 2013, p.
        16).

        Adult Recommendation (18 years and older):

        'The USPSTF recommends screening for depression in the general adult population, including
        pregnant and postpartum women. Screening should be implemented with adequate systems in
        place to ensure accurate diagnosis, effective treatment, and appropriate follow-up (B
        recommendation)' (Sui, A. and USPSTF, 2016, p. 380).

        The Institute for Clinical Systems Improvement (ICSI) health care guideline, Adult
        Depression in Primary Care, provides the following recommendations:
        1. 'Clinicians should routinely screen all adults for depression using a standardized
        instrument.'
        2. 'Clinicians should establish and maintain follow-up with patients.'
        3. 'Clinicians should screen and monitor depression in pregnant and post-partum women.'
        (Trangle, 2016 p.p. 9 - 10)
        """
        patient = self.patient
        age = self.patient.age_at(self.timeframe.end)

        recommendations: List[Recommendation] = []
        status = 'not_applicable'
        narrative = ''

        # FIX: is looking back one year correct?  The measure doesn't
        # say how often you should screen your patients.
        screening = (self.patient.interviews
                     .find(AdolescentDepressionScreening | AdultDepressionScreening)
                     .within_one_year(self.timeframe)
                     .last())  # yapf: disable

        if not screening:
            # should get a screening
            rec = InterviewRecommendation(
                patient=patient,
                questionnaire=(AdultDepressionScreening
                               if age >= 18 else AdultDepressionScreening))
            recommendations.append(rec)

            status = 'due'
            narrative = rec.narrative
            return {'status': status, 'narrative': narrative, 'recommendations': recommendations}

        # if they got the screening and they're negative we're done
        if 'positive' not in screening['results'][0]['narrative'].lower():
            status = 'current'
            narrative = f'{patient.first_name} was screened for depression with negative results.'
            return {'status': status, 'narrative': narrative, 'recommendations': recommendations}

        # they're screened and positive -- if they've gotten treatment then we're done
        if self.patient.instructions.find(FollowUpForDepressionAdolescent |
                                          FollowUpForDepressionAdult).after(
                                              arrow.get(screening['created'])):
            status = 'current'
            narrative = f'{patient.first_name} was screened for depression with positive results and a followup '\
                         'is documented.'
            return {'status': status, 'narrative': narrative, 'recommendations': recommendations}

        # recommend a followup
        recommendations.append(
            InstructionRecommendation(
                patient=patient,
                instruction=(FollowUpForDepressionAdult
                             if age >= 18 else FollowUpForDepressionAdolescent)))
        status = 'due'
        narrative = f'{patient.first_name} was screened for depression with positive results, a followup is recommended.'

        return {'status': status, 'narrative': narrative, 'recommendations': recommendations}



from unittest.mock import PropertyMock, patch

import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.ccp002v1_high_risk_outreach import MINIMUM_AGE, Ccp002v1
from canvas_workflow_kit.patient_recordset import ConditionRecordSet, InterviewRecordSet
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class TestCcp002v1(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_minimal_age(self):
        assert 65 == MINIMUM_AGE

    def test_enabled(self):
        self.assertTrue(Ccp002v1.enabled())

    def test_description(self):
        expected = 'All patients with 65+ with chronic conditions to be reached.'
        self.assertEqual(expected, Ccp002v1._meta.description)

    def test_information(self):
        expected = (
            'https://canvas-medical.zendesk.com/hc/en-us/articles/360059084173-COVID-19-Risk-Assessment-Follow-Up-Protocol')
        self.assertEqual(expected, Ccp002v1._meta.information)

    def test_change_types(self):
        result = Ccp002v1._meta.compute_on_change_types
        expected = ['condition', 'interview', 'patient']
        self.assertEqual(expected, result)

    def test_version(self):
        self.assertTrue(hasattr(Ccp002v1._meta, 'version'))

    def test_date_of_first_questionnaire(self):
        start = arrow.get('2019-03-24 13:24:56')
        end = arrow.get('2020-03-24 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # no interview
        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.date_of_first_questionnaire
        self.assertEqual('', result)

        # interview symptomatic surveillance
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire('2020-03-14 13:24:55+00:00', 'CANVAS0001')])
        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.date_of_first_questionnaire
        self.assertEqual('2020-03-14 13:24:55+00:00', result)

        # interview high risk outreach
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire('2020-03-14 13:24:55+00:00', 'CANVAS0006')])
        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.date_of_first_questionnaire
        self.assertEqual('2020-03-14 13:24:55+00:00', result)

        # interviews -> the oldest
        patient.interviews = InterviewRecordSet([
            self.helper_questionnaire('2020-03-14 13:24:55+00:00', 'CANVAS0001'),
            self.helper_questionnaire('2020-03-14 13:24:54+00:00', 'CANVAS0006'),
            self.helper_questionnaire('2020-03-14 13:24:56+00:00', 'CANVAS0001'),
        ])
        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.date_of_first_questionnaire
        self.assertEqual('2020-03-14 13:24:54+00:00', result)

    def test_high_risk_conditions(self):
        start = arrow.get('2019-03-24 13:24:56')
        end = arrow.get('2020-03-24 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # no condition -> False
        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.high_risk_conditions
        self.assertEqual([], result)

        # not high risk condition -> False
        patient.conditions = ConditionRecordSet([self.helper_condition('active', '0QPD0JZ')])
        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.high_risk_conditions
        self.assertEqual([], result)

        # high risk condition --> active > True, resolved > False
        tests = [
            'K7150',
            'J441',
            'E119',
            'C8122',
            'I63531',
            'N172',
            'Z6841',
            'J4532',
        ]
        for code in tests:
            patient.conditions = ConditionRecordSet([self.helper_condition('active', code)])
            tested = Ccp002v1(patient=patient, timeframe=timeframe)
            result = tested.high_risk_conditions
            self.assertEqual([{
                'clinicalStatus': 'active',
                'coding': [{
                    'code': code,
                    'system': 'ICD-10'
                }],
                'periods': [{
                    'from': '2018-08-20',
                    'to': None
                }],
            }], result)

            patient.conditions = ConditionRecordSet([self.helper_condition('resolved', code)])
            tested = Ccp002v1(patient=patient, timeframe=timeframe)
            result = tested.high_risk_conditions
            self.assertEqual([], result)

        # several high risk conditions --> list all
        patient.conditions = ConditionRecordSet([
            self.helper_condition('active', 'E119'),
            self.helper_condition('active', 'Z6841'),
        ])
        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.high_risk_conditions
        self.assertEqual([
            {
                'clinicalStatus': 'active',
                'coding': [{
                    'code': 'E119',
                    'system': 'ICD-10'
                }],
                'periods': [{
                    'from': '2018-08-20',
                    'to': None
                }],
            },
            {
                'clinicalStatus': 'active',
                'coding': [{
                    'code': 'Z6841',
                    'system': 'ICD-10'
                }],
                'periods': [{
                    'from': '2018-08-20',
                    'to': None
                }],
            },
        ], result)

    def test_in_initial_population(self):
        with patch('canvas_workflow_kit.builtin_cqms.ccp002v1_high_risk_outreach.arrow.now') as now:
            now.return_value = arrow.get('2020-03-22 13:24:56')

            start = arrow.get('2019-03-22 13:24:56')
            end = arrow.get('2020-03-22 13:24:56')
            timeframe = Timeframe(start=start, end=end)
            patient = self.load_patient('patient')

            # 65+ + no high risk condition --> True
            patient.patient['birthDate'] = '1955-03-20'
            patient.conditions = ConditionRecordSet([self.helper_condition('active', 'H3533')])
            tested = Ccp002v1(patient=patient, timeframe=timeframe)
            result = tested.in_initial_population()
            self.assertTrue(result)

            # <65 + high risk condition --> True
            patient.patient['birthDate'] = '1955-03-20'
            patient.conditions = ConditionRecordSet([self.helper_condition('active', 'C8122')])
            tested = Ccp002v1(patient=patient, timeframe=timeframe)
            result = tested.in_initial_population()
            self.assertTrue(result)

            # <65 +  no high risk condition -> False
            patient.patient['birthDate'] = '1955-03-24'
            patient.conditions = ConditionRecordSet([self.helper_condition('active', 'H3533')])
            tested = Ccp002v1(patient=patient, timeframe=timeframe)
            result = tested.in_initial_population()
            self.assertFalse(result)

    def test_in_denominator(self):
        with patch('canvas_workflow_kit.builtin_cqms.ccp002v1_high_risk_outreach.arrow.now') as now:
            now.return_value = arrow.get('2020-03-22 13:24:56')

            start = arrow.get('2019-03-22 13:24:56')
            end = arrow.get('2020-03-22 13:24:56')
            timeframe = Timeframe(start=start, end=end)
            patient = self.load_patient('patient')

            # 65+ + no high risk condition --> True
            patient.patient['birthDate'] = '1955-03-21'
            patient.conditions = ConditionRecordSet([self.helper_condition('active', 'H3533')])
            tested = Ccp002v1(patient=patient, timeframe=timeframe)
            result = tested.in_denominator()
            self.assertTrue(result)

            # <65 + high risk condition --> True
            patient.patient['birthDate'] = '1955-03-24'
            patient.conditions = ConditionRecordSet([self.helper_condition('active', 'C8122')])
            tested = Ccp002v1(patient=patient, timeframe=timeframe)
            result = tested.in_denominator()
            self.assertTrue(result)

            # <65 + no high risk condition -> False
            patient.conditions = ConditionRecordSet([self.helper_condition('active', 'H3533')])
            tested = Ccp002v1(patient=patient, timeframe=timeframe)
            result = tested.in_denominator()
            self.assertFalse(result)

    def test_in_numerator(self):
        start = arrow.get('2019-03-24 13:24:56')
        end = arrow.get('2020-03-24 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # no interview
        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.in_numerator()
        self.assertFalse(result)

        # interview not Covid-19 -> False
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire('2020-03-14 13:24:55+00:00', 'CANVAS0002')])
        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.in_numerator()
        self.assertFalse(result)

        # interview symptomatic surveillance -> True
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire('2020-03-14 13:24:55+00:00', 'CANVAS0001')])
        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.in_numerator()
        self.assertTrue(result)

        # interview high risk outreach -> True
        patient.interviews = InterviewRecordSet(
            [self.helper_questionnaire('2020-03-14 13:24:55+00:00', 'CANVAS0006')])
        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.in_numerator()
        self.assertTrue(result)

        # multiple including covid-19 -> True
        patient.interviews = InterviewRecordSet([
            self.helper_questionnaire('2020-03-14 13:24:55+00:00', 'CANVAS0001'),
            self.helper_questionnaire('2020-03-14 13:24:54+00:00', 'CANVAS0006'),
            self.helper_questionnaire('2020-03-14 13:24:56+00:00', 'CANVAS0002'),
        ])
        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.in_numerator()
        self.assertTrue(result)

    @patch(
        'canvas_workflow_kit.builtin_cqms.ccp002v1_high_risk_outreach.Ccp002v1.high_risk_conditions',
        new_callable=PropertyMock)
    @patch(
        'canvas_workflow_kit.builtin_cqms.ccp002v1_high_risk_outreach.Ccp002v1.date_of_first_questionnaire',
        new_callable=PropertyMock)
    @patch('canvas_workflow_kit.builtin_cqms.ccp002v1_high_risk_outreach.Ccp002v1.in_numerator')
    @patch('canvas_workflow_kit.builtin_cqms.ccp002v1_high_risk_outreach.Ccp002v1.in_denominator')
    @patch('canvas_workflow_kit.builtin_cqms.ccp002v1_high_risk_outreach.arrow.now')
    def test_compute_results(self, now, in_denominator, in_numerator, date_of_first_questionnaire,
                             high_risk_conditions):
        now.return_value = arrow.get('2020-03-22 13:24:56')

        start = arrow.get('2019-03-24 13:24:56')
        end = arrow.get('2020-03-24 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('patient')

        # not in_denominator -> not_applicable
        high_risk_conditions.return_value = []
        in_denominator.return_value = False
        in_numerator.return_value = True
        date_of_first_questionnaire.return_value = '2020-03-14 13:24:55+00:00'

        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('not_applicable', result.status)
        self.assertIsNone(result.due_in)

        # in_denominator + in_numerator -> satisfied
        high_risk_conditions.return_value = []
        in_denominator.return_value = True
        in_numerator.return_value = True
        date_of_first_questionnaire.return_value = '2020-03-14 13:24:55+00:00'

        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        self.assertEqual(-1, result.due_in)
        narratives = ['Nicolas took a COVID-19 outreach questionnaire on Sat, Mar 14th 2020.']
        self.assertEqual(narratives, result.narratives)
        self.assertEqual(0, len(result.recommendations))

        # patient is less than 65
        patient.patient['birthDate'] = '1956-03-20'
        # in_denominator + not in_numerator + no display -> due + count of high risk conditions
        high_risk_conditions.return_value = [
            {
                'clinicalStatus': 'active',
                'coding': [{
                    'code': 'E119',
                    'system': 'ICD-10'
                }],
                'periods': [{
                    'from': '2018-08-20',
                    'to': None
                }],
            },
            {
                'clinicalStatus': 'active',
                'coding': [{
                    'code': 'Z6841',
                    'system': 'ICD-10'
                }],
                'periods': [{
                    'from': '2018-08-20',
                    'to': None
                }],
            },
        ]
        in_denominator.return_value = True
        in_numerator.return_value = False
        date_of_first_questionnaire.return_value = '2020-03-14 13:24:55+00:00'

        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        self.assertEqual(0, result.due_in)
        narratives = [
            'Nicolas should take a COVID-19 outreach questionnaire.',
            'Nicolas has 2 high risk conditions',
        ]
        self.assertEqual(narratives, result.narratives)
        self.assertEqual(1, len(result.recommendations))
        recommendation = result.recommendations[0]
        self.assertEqual('CCP002v1_RECOMMEND_QUESTIONNAIRE', recommendation.key)
        self.assertEqual('Complete COVID-19 outreach questionnaire', recommendation.title)

        # in_denominator + not in_numerator + display -> due + list of high risk conditions
        high_risk_conditions.return_value = [
            {
                'clinicalStatus': 'active',
                'coding': [{
                    'code': 'E119',
                    'system': 'ICD-10',
                    'display': 'diabetes'
                }],
                'periods': [{
                    'from': '2018-08-20',
                    'to': None
                }],
            },
            {
                'clinicalStatus': 'active',
                'coding': [{
                    'code': 'Z6841',
                    'system': 'ICD-10',
                    'display': 'body mass'
                }],
                'periods': [{
                    'from': '2018-08-20',
                    'to': None
                }],
            },
        ]
        in_denominator.return_value = True
        in_numerator.return_value = False
        date_of_first_questionnaire.return_value = '2020-03-14 13:24:55+00:00'

        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        self.assertEqual(0, result.due_in)
        narratives = [
            'Nicolas should take a COVID-19 outreach questionnaire.',
            'Nicolas has high risk conditions:',
            'diabetes',
            'body mass',
        ]
        self.assertEqual(narratives, result.narratives)
        self.assertEqual(1, len(result.recommendations))
        recommendation = result.recommendations[0]
        self.assertEqual('CCP002v1_RECOMMEND_QUESTIONNAIRE', recommendation.key)
        self.assertEqual('Complete COVID-19 outreach questionnaire', recommendation.title)

        # patient is more than 65
        patient.patient['birthDate'] = '1954-03-20'
        # in_denominator + not in_numerator + display -> due + list of high risk conditions
        high_risk_conditions.return_value = [
            {
                'clinicalStatus': 'active',
                'coding': [{
                    'code': 'E119',
                    'system': 'ICD-10',
                    'display': 'diabetes'
                }],
                'periods': [{
                    'from': '2018-08-20',
                    'to': None
                }],
            },
            {
                'clinicalStatus': 'active',
                'coding': [{
                    'code': 'Z6841',
                    'system': 'ICD-10',
                    'display': 'body mass'
                }],
                'periods': [{
                    'from': '2018-08-20',
                    'to': None
                }],
            },
        ]
        in_denominator.return_value = True
        in_numerator.return_value = False
        date_of_first_questionnaire.return_value = '2020-03-14 13:24:55+00:00'

        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        self.assertEqual(0, result.due_in)
        narratives = [
            'Nicolas should take a COVID-19 outreach questionnaire.',
            'Nicolas is 65+',
            'Nicolas has high risk conditions:',
            'diabetes',
            'body mass',
        ]
        self.assertEqual(narratives, result.narratives)
        self.assertEqual(1, len(result.recommendations))
        recommendation = result.recommendations[0]
        self.assertEqual('CCP002v1_RECOMMEND_QUESTIONNAIRE', recommendation.key)
        self.assertEqual('Complete COVID-19 outreach questionnaire', recommendation.title)

        # in_denominator + not in_numerator + no high risk condition -> due + age
        high_risk_conditions.return_value = []
        in_denominator.return_value = True
        in_numerator.return_value = False
        date_of_first_questionnaire.return_value = '2020-03-14 13:24:55+00:00'

        tested = Ccp002v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        self.assertEqual(0, result.due_in)
        narratives = [
            'Nicolas should take a COVID-19 outreach questionnaire.',
            'Nicolas is 65+',
        ]
        self.assertEqual(narratives, result.narratives)
        self.assertEqual(1, len(result.recommendations))
        recommendation = result.recommendations[0]
        self.assertEqual('CCP002v1_RECOMMEND_QUESTIONNAIRE', recommendation.key)
        self.assertEqual('Complete COVID-19 outreach questionnaire', recommendation.title)

    def helper_condition(self, status: str, icd10: str):
        return {
            'clinicalStatus': status,
            'coding': [{
                'code': icd10,
                'system': 'ICD-10'
            }],
            'periods': [{
                'from': '2018-08-20',
                'to': None if status == 'active' else '2020-03-21'
            }],
        }

    def helper_questionnaire(self, date: str, canvas: str):
        return {
            'noteTimestamp': date,
            'name': 'COVID Questionnaire',
            'results': [],
            'questionnaires': [{
                'code': canvas,
                'codeSystem': 'CANVAS'
            }],
        }



from typing import List

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import ClinicalQualityMeasure
from canvas_workflow_kit.recommendation import Recommendation

# flake8: noqa
from canvas_workflow_kit.value_set.v2018 import (
    ConsultantReport, Ethnicity, FaceToFaceInteraction, OfficeVisit, OncAdministrativeSex,
    OphthalmologicalServices, Payer, PreventiveCareEstablishedOfficeVisit0To17,
    PreventiveCareInitialOfficeVisit0To17, PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit18AndUp, Race, Referral)


class ClinicalQualityMeasure50v6(ClinicalQualityMeasure):
    """
    Closing the Referral Loop: Receipt of Specialist Report

    Description: Percentage of patients with referrals, regardless of age, for which the referring
    provider receives a report from the provider to whom the patient was referred

    Definition: Referral: A request from one physician or other eligible provider to another
    practitioner for evaluation, treatment, or co-management of a patient's condition. This term
    encompasses referral and consultation as defined by Centers for Medicare and Medicaid Services.

    Rationale: Problems in the outpatient referral and consultation process have been documented,
    including lack of timeliness of information and inadequate provision of information between the
    specialist and the requesting physician (Gandhi, 2000; Forrest, 2000; Stille, 2005). In a study
    of physician satisfaction with the outpatient referral process, Gandhi et al. (2000) found that
    68% of specialists reported receiving no information from the primary care provider prior to
    referral visits, and 25% of primary care providers had still not received any information from
    specialists 4 weeks after referral visits. In another study of 963 referrals (Forrest, 2000),
    pediatricians scheduled appointments with specialists for only 39% and sent patient information
    to the specialists in only 51% of the time.

    In a 2006 report to Congress, MedPAC found that care coordination programs improved quality of
    care for patients, reduced hospitalizations, and improved adherence to evidence-based care
    guidelines, especially among patients with diabetes and CHD. Associations with cost-savings
    were less clear; this was attributed to how well the intervention group was chosen and defined,
    as well as the intervention put in place. Additionally, cost-savings were usually calculated in
    the short-term, while some argue that the greatest cost-savings accrue over time (MedPAC,
    2006).

    Improved mechanisms for information exchange could facilitate communication between providers,
    whether for time-limited referrals or consultations, on-going co-management, or during care
    transitions. For example, a study by Branger et al. (1999) found that an electronic
    communication network that linked the computer-based patient records of physicians who had
    shared care of patients with diabetes significantly increased frequency of communications
    between physicians and availability of important clinical data. There was a 3-fold increase in
    the likelihood that the specialist provided written communication of results if the primary
    care physician scheduled appointments and sent patient information to the specialist (Forrest,
    2000).

    Care coordination is a focal point in the current health care reform and our nation's
    ambulatory health information technology (HIT) framework. The National Priorities Partnership
    recently highlighted care coordination as one of the most critical areas for development of
    quality measurement and improvement (NPP, 2008).

    Guidance: The provider who refers the patient to another provider is the provider who should be
    held accountable for the performance of this measure.

    The provider to whom the patient was referred should be the same provider that sends the
    report.

    If there are multiple referrals for a patient during the measurement period, use the first
    referral.

    The consultant report that will fulfill the referral should be completed after the referral,
    and should be related to the referral for which it is attributed. If there are multiple
    consultant reports received by the referring provider which pertain to a particular referral,
    use the first consultant report to satisfy the measure. Eligible professionals or eligible
    clinicians reporting on this measure should note that all data for the reporting year is to be
    submitted by the deadline established by CMS. Therefore, eligible professionals or eligible
    clinicians who see patients towards the end of the reporting period (ie, December in
    particular), should communicate the consultant report as soon as possible in order for those
    patients to be counted in the measure numerator. Communicating the report as soon as possible
    will ensure the data is included in the submission to CMS.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms50v6
    """

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'Branger, P. J., Van\'t Hooft, A., Van Der Wouden, J. C., Moorman, P. W., and Van Bemmel, J. H. (1999). Shared care for diabetes: supporting communication between primary and secondary care. International Journal of Medical Informatics 53(2-3), 133-142.',
        'Forrest, C. B., Glade, G. B., Baker, A. E., Bocian, A., Von Schrader, S., and Starfield, B. (2000). Coordination of specialty referrals and physician satisfaction with referral care. Archives of Pediatrics and Adolescent Medicine 154(5), 499-506.',
        'Gandhi, T. K., Sittig, D. F., Franklin, M., Sussman, A. J., Fairchild, D. G., and Bates, D. W. (2000). Communication breakdown in the outpatient referral process. Journal of General Internal Medicine 15(9), 626-631.',
        'Medicare Payment Advisory Commission (MedPAC) Report to the Congress: Medicare Payment Policy.March, 2006. Retrieved February 22, 2017, from http://medpac.gov/docs/default-source/reports/Mar06_EntireReport.pdf?sfvrsn=0',
        'National Priorities Partnership. National Priorities and Goals: Aligning Our Efforts to Transform America\'s Healthcare. Washington, DC: National Quality Forum; 2008.',
        'Stille, C. J., Jerant, A., Bell, D., Meltzer, D., and Elmore, J. G. (2005). Coordinating care across diseases, settings, and clinicians: a key role for the generalist in practice. Annals of Internal Medicine 142(8), 700-708.',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Number of patients, regardless of age, who were referred by one
        provider to another provider, and who had a visit during the measurement period
        """
        if self.patient.referrals.within_one_year(self.timeframe):
            return True

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: None

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Number of patients with a referral, for which the referring provider received a
        report from the provider to whom the patient was referred

        Exclusions: Not Applicable
        """
        # FIX: how do we find reports and match them back to referrals?

        return False

    def compute_results(self):
        """
        Clinical recommendation: None
        """
        # FIX: how do we find reports and match them back to referrals?

        recommendations: List[Recommendation] = []
        status = 'not_applicable'
        narrative = ''

        return {'status': status, 'narrative': narrative, 'recommendations': recommendations}



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    AcuteInpatient, AnnualWellnessVisit, CareServicesInLongTermResidentialFacility,
    DementiaMedications, Ed, EncounterInpatient, Ethnicity, Female, FrailtyDevice,
    FrailtyDiagnosis, FrailtyEncounter, FrailtySymptom, HistoryOfBilateralMastectomy,
    HomeHealthcareServices, HospiceCareAmbulatory, Mammography, NonacuteInpatient,
    NursingFacilityVisit, Observation, OfficeVisit, OncAdministrativeSex, Outpatient, Payer,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit18AndUp, Race, StatusPostLeftMastectomy,
    StatusPostRightMastectomy, UnilateralMastectomyUnspecifiedLaterality)


class ClinicalQualityMeasure125v9(ClinicalQualityMeasure):
    """
    Breast Cancer Screening

    Description: Percentage of women 50-74 years of age who had a mammogram to screen for breast
    cancer in the 27 months prior to the end of the Measurement Period

    Definition: None

    Rationale: Breast cancer is one of the most common types of cancers, accounting for 15 percent
    of all new cancer diagnoses in the U.S. (Noone et al, 2018). In 2015, over 3 million women were
    estimated to be living with breast cancer in the U.S. and it is estimated that 12 percent of
    women will be diagnosed with breast cancer at some point during their lifetime (Noone et al,
    2018).

    While there are other factors that affect a woman's risk of developing breast cancer, advancing
    age is a primary risk factor. Breast cancer is most frequently diagnosed among women ages
    55-64; the median age at diagnosis is 62 years (Noone et al, 2018).

    The chance of a woman being diagnosed with breast cancer in a given year increases with age. By
    age 40, the chances are 1 in 68; by age 50 it becomes 1 in 43; by age 60, it is 1 in 29
    (American Cancer Society, 2017).

    Guidance: Patient self-report for procedures as well as diagnostic studies should be recorded
    in 'Procedure, Performed' template or 'Diagnostic Study, Performed' template in QRDA-1.

    This measure evaluates primary screening. Do not count biopsies, breast ultrasounds, or MRIs
    because they are not appropriate methods for primary breast cancer screening.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms125v9
    """

    title = 'Breast Cancer Screening'

    identifiers = ['CMS125v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'American Cancer Society. 2017. “Breast Cancer Facts & Figures 2017-2018.” (February 8, 2019). Retrieved from https://www.cancer.org/content/dam/cancer-org/research/cancer-facts-and-statistics/breast-cancer-facts-and-figures/breast-cancer-facts-and-figures-2017-2018.pdf',
        'Noone, A.M., Howlader, N., Krapcho, M., Miller, D., Brest, A., Yu, M., Ruhl, J., Tatalovich, Z., Mariotto, A., Lewis, D.R., Chen, H.S., Feuer, E.J., Cronin, K.A. (eds). 2018. “SEER Cancer Statistics Review, 1975-2015.” National Cancer Institute. Bethesda, MD. (February 8, 2019) Retrieved from https://seer.cancer.gov/csr/1975_2015/',
        'U.S. Preventive Services Task Force. (2016). Screening for breast cancer: U.S. Preventive Services Task Force recommendation statement. Annals of Internal Medicine, 164(4), 279-296.',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: Women 51-74 years of age with a visit during the measurement period
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: Women who had a bilateral mastectomy or who have a history of a bilateral
        mastectomy or for whom there is evidence of a right and a left unilateral mastectomy.

        Exclude patients whose hospice care overlaps the measurement period.

        Exclude patients 66 and older who are living long term in an institution for more than 90
        consecutive days during the measurement period.

        Exclude patients 66 and older with advanced illness and frailty because it is unlikely that
        patients will benefit from the services being measured.

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Women with one or more mammograms during the 27 months prior to the end of the
        measurement period

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: The U.S. Preventive Services Task Force (USPSTF) recommends
        biennial screening mammography for women aged 50-74 years (B recommendation).

        The decision to start screening mammography in women prior to age 50 years should be an
        individual one. Women who place a higher value on the potential benefit than the potential
        harms may choose to begin biennial screening between the ages of 40 and 49 years (C
        recommendation). (USPSTF, 2016)

        The USPSTF concludes that the current evidence is insufficient to assess the balance of
        benefits and harms of screening mammography in women aged 75 years or older (I statement).
        (USPSTF, 2016)

        The USPSTF concludes that the current evidence is insufficient to assess the benefits and
        harms of digital breast tomosynthesis (DBT) as a primary screening method for breast cancer
        (I Statement). (USPSTF, 2016)

        The USPSTF concludes that the current evidence is insufficient to assess the balance of
        benefits and harms of adjunctive screening for breast cancer using breast ultrasonography,
        magnetic resonance imaging, DBT, or other methods in women identified to have dense breasts
        on an otherwise negative screening mammogram (I statement). (USPSTF, 2016)
        """
        pass



# flake8: noqa
from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import ClinicalQualityMeasure
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.value_set.v2017 import (
    AcuteMyocardialInfarction, AnnualWellnessVisit, CoronaryArteryBypassGraft, Ethnicity,
    FaceToFaceInteraction, HdlCLaboratoryTest, HomeHealthcareServices, IschemicVascularDisease,
    LdlC, OfficeVisit, OncAdministrativeSex, Payer, PercutaneousCoronaryInterventions,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit18AndUp, Race, TotalCholesterolLaboratoryTest,
    TriglyceridesLaboratoryTest)


# TODO Recommendations can have >1 action (more than one way to resolve a Recommendation)
# TODO Add Predicate class for rich "debugging"?
# TODO action-oriented 'tell me what to do view' vs. inverted 'tell me why' view
# TODO due/pending/etc. based on order committed/status
class ClinicalQualityMeasure182v6(ClinicalQualityMeasure):
    """
    Ischemic Vascular Disease (IVD): Complete Lipid Profile and LDL-C Control (<100 mg/dL)

    Description: Percentage of patients 18 years of age and older who were diagnosed with acute
    myocardial infarction (AMI), coronary artery bypass graft (CABG) or percutaneous coronary
    interventions (PCI) in the 12 months prior to the measurement period, or who had a diagnosis of
    ischemic vascular disease (IVD) during the measurement period, and who had a complete lipid
    profile performed during the measurement period and whose most recent Low-density Lipoprotein
    (LDL-C) was adequately controlled (< 100 mg/dL)

    Definition: None

    Rationale: A 10 percent decrease in total cholesterol levels (population wide) may result in an
    estimated 30 percent reduction in the incidence of coronary heart disease (CHD) (Centers for
    Disease Control and Prevention 2000). Based on data from the Third Report of the Expert Panel
    on Detection, Evaluation, and Treatment of High Blood Cholesterol in Adults:
    *Less than half of persons who qualify for any kind of lipid-modifying treatment for CHD risk
    reduction are receiving it
    *Less than half of even the highest-risk persons, those who have symptomatic CHD, are receiving
    lipid-lowering treatment
    *Only about a third of treated patients are achieving their LDL goal; less than 20 percent of
    CHD patients are at their LDL goal (National Cholesterol Education Program (NCEP) Expert Panel
    on Detection, Evaluation, and Treatment of High Blood Pressure 2002)

    According to data from the Behavioral Risk Factor Surveillance System (BRFSS) from 1991 - 2003,
    the prevalence of cholesterol screening during the preceding 5 years increased from 67.3
    percent in 1991 to 73.1 percent in 2003 (Centers for Disease Control and Prevention 2005).

    Between 1988-94 and 1999-2002, the age-adjusted mean total serum cholesterol level of adults 20
    years of age and  older decreased from 206 mg/dL to 203 mg/dL, and LDL cholesterol levels
    decreased from 129 mg/dL to 123 mg/dL. The mean level of LDL cholesterol for American adults
    age 20 and older is 123 mg/dL (Carroll et al. 2005). However, even given this decrease, there
    is still a significant amount of room for improvement.

    Guidance: Only patients who were diagnosed with acute myocardial infarction (AMI), coronary
    artery bypass graft (CABG) or percutaneous coronary interventions (PCI) are included in this
    measure

    More information: https://ecqi.healthit.gov/ecqm/measures/cms182v6
    """

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'National Committee for Quality Assurance',
    ]

    references = [
        'Centers for Disease Control and Prevention (CDC). 2005. "Trends in cholesterol screening and awareness of high blood cholesterol - United States, 1991-2003." MMWR 54;865-870.',
        'Carroll, M.D., D.A. Lacher, P.D. Sorlie, J.I. Cleeman, D.J. Gordon, M. Wolz, S.M. Grundy, C.L. Johnson. 2005. "Trends in serum lipids and lipoproteins of adults. 1960-2002." JAMA 294:1773-1781.',
        'Centers for Disease Control and Prevention (CDC). 2000. "State-specific cholesterol screening trends - United States, 1991-1999." MMWR 49:750-755.',
        'National Cholesterol Education Program (NCEP) Expert Panel on Detection, Evaluation, and Treatment of High Blood Cholesterol in Adults (Adult Treatment Panel III). 2002. "Third Report of the National Cholesterol Education Program (NCEP) Expert Panel on Detection, Evaluation, and Treatment of High Blood Cholesterol in Adults (Adult Treatment Panel III) final report." Circulation 106(25):3143-421.',
        'Grundy SM, Cleeman JI, Merz CN, et al. Implications of recent clinical trials for the National Cholesterol Education Program Adult Treatment Panel III guidelines. Circulation 2004; 110: 227-39.',
        'U.S. Preventive Services Task Force. 2008. "Screening for lipid disorders in adults" (June) http://www.uspreventiveservicestaskforce.org/uspstf/uspschol.htm',
    ]

    funding_source = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.one_year_before_start = Timeframe(
            start=self.timeframe.start.shift(months=-12), end=self.timeframe.start)

    def has_acute_myocardial_infarction(self):
        return self.patient.conditions.within(
            self.one_year_before_start).find(AcuteMyocardialInfarction)

    def has_ischemic_vascular_disease(self):
        return self.patient.conditions.within(self.timeframe).find(IschemicVascularDisease)

    def has_percutaneous_coronary_interventions(self):
        return self.patient.conditions.within(
            self.one_year_before_start).find(PercutaneousCoronaryInterventions)

    def has_coronary_artery_bypass_graft(self):
        return self.patient.conditions.within(
            self.one_year_before_start).find(CoronaryArteryBypassGraft)

    def has_all_four_tests_during_measurement_period(self):
        # XXX is checking for a result within the timeframe sufficient?
        return all([
            self.patient.lab_results.within(self.timeframe).find(HdlCLaboratoryTest),
            self.patient.lab_results.within(self.timeframe).find(LdlC),
            self.patient.lab_results.within(self.timeframe).find(TotalCholesterolLaboratoryTest),
            self.patient.lab_results.within(self.timeframe).find(TriglyceridesLaboratoryTest),
        ])

    def has_ldl_below_100mg_dl(self):
        most_recent_ldl_c = self.patient.lab_results.within(
            self.timeframe).find(LdlC).last_value(float)

        return most_recent_ldl_c is not None and float(most_recent_ldl_c['value']) < 100

    def in_initial_population(self):
        """
        Initial population: Patients 18 years of age and older with a visit during the measurement
        year who had an AMI, CABG, or PCI during the 12 months prior to the measurement year or who
        had a diagnosis of IVD during the measurement year
        """
        return any([
            self.has_acute_myocardial_infarction(),
            self.has_ischemic_vascular_disease(),
            all([
                self.has_percutaneous_coronary_interventions(),
                self.has_coronary_artery_bypass_graft(),
            ]),
        ])

    def in_denominator(self):
        """
        Denominator: Equals Initial Population
        """
        return self.in_initial_population()

    def in_numerator_1(self):
        """
        Numerator 1: Patients with a complete lipid profile performed during the measurement period
        """
        return self.has_all_four_tests_during_measurement_period()

    def in_numerator_2(self):
        """
        Numerator 2: Patients whose most recent LDL-C level performed during the measurement period
        is <100 mg/dL
        """
        return self.has_ldl_below_100mg_dl()

    def recommendations(self):
        """
        Clinical recommendation: Third report of the National Cholesterol Education Program (NCEP)
        Expert Panel on Detection, Evaluation, and Treatment of High Blood Cholesterol in Adults
        (Adult Treatment Panel III). (2002) AND Implications of recent clinical trials for the
        National Cholesterol Education Program Adult Treatment Panel III guidelines (2004)
        In high-risk persons, the recommended LDL-C goal is < 100 mg/dL.

        * An LDL-C goal of < 70 mg/dL is a therapeutic option on the basis of available clinical
          trial evidence, especially for patients at very high risk.

        * If LDL-C is > 100 mg/dL, an LDL-lowering drug is indicated simultaneously with lifestyle
          changes.

        * If baseline LDL-C is < 100 mg/dL, institution of an LDL-lowering drug to achieve an LDL-C
          level < 70 mg/dL is a therapeutic option on the basis of available clinical trial
          evidence.

        * If a high-risk person has high triglycerides or low HDL-C, consideration can be given to
          combining a fibrate or nicotinic acid with an LDL-lowering drug. When triglycerides are >
          200 mg/dL, non-HDL-C is a secondary target of therapy, with a goal 30 mg/dL higher than
          the identified LDL-C goal.

        The U.S. Preventive Services Task Force (USPSTF) strongly recommends screening men aged 35
        and older for lipid disorders and recommends screening men aged 20 to 35 for lipid
        disorders if they are at increased risk for coronary heart disease. The USPSTF also
        strongly recommends screening women aged 45 and older for lipid disorders if they are at
        increased risk for coronary heart disease and recommends screening women aged 20 to 45 for
        lipid disorders if they are at increased risk for coronary heart disease.
        """
        pass



"""

Recommendations are classes that provide guidance
toward the next course of action.

** Example: **
```python
def compute_results(self) -> ProtocolResult:
    result = ProtocolResult()

    if self.has_suspect_eye_condition:
        title = ('Consider updating the Diabetes without complications (E11.9) '
                 'to Diabetes with secondary eye disease as clinically appropriate.')
        result.add_narrative(comment)
        result.add_diagnose_recommendation(
            key='HCC003v1_RECOMMEND_DIAGNOSE_EYE',
            rank=1,
            button='Diagnose',
            title=title,
            narrative=f'{self.patient.first_name} has Diabetes without complications,
            command={'key': 'diagnose'})

"""

from typing import List, Optional, Type, Dict

from .constants import (
    COMMAND_TYPE_ALLERGY,
    COMMAND_TYPE_DIAGNOSE,
    COMMAND_TYPE_FOLLOWUP,
    COMMAND_TYPE_IMAGING,
    COMMAND_TYPE_IMMUNIZATION,
    COMMAND_TYPE_INSTRUCTION,
    COMMAND_TYPE_INTERVIEW,
    COMMAND_TYPE_LAB_ORDER,
    COMMAND_TYPE_PERFORM,
    COMMAND_TYPE_PLAN,
    COMMAND_TYPE_PRESCRIBE,
    COMMAND_TYPE_REFERRAL,
    COMMAND_TYPE_STRUCTURED_ASSESSMENT,
    COMMAND_TYPE_TASK,
    COMMAND_TYPE_VITALSIGN
)
from .patient import Patient
from .value_set.value_set import ValueSet
from .intervention import Intervention

# Lower numbers displayed first
RANK_MAX = 1000


class Recommendation(Intervention):
    """
    A standard recommendation.
    """

    # HACK: title and narrative have defaults here so I didn't have to reorder the arguments when
    # adding a default for rank (to make it optional)
    def __init__(self,
                 key,
                 rank=RANK_MAX,
                 button='',
                 title=None,
                 narrative=None,
                 command=None,
                 context=None,
                 href=None):
        if not title:
            raise ValueError

        self.button = button
        self.command = command or {}
        self.context = context or {}
        self.key = key
        self.narrative = narrative
        self.rank = rank
        self.title = title
        self.href = href

    @staticmethod
    def command_filter(command_type: str, value_sets: List[Type[ValueSet]]):
        """
        Build up a command that contains a search filter by value set
        system codes, like a search for a particular vaccine or lab or
        medicine to prescribe.
        """

        # sets to lists for json encode - can't directly encode sets
        codes = {}
        for value_set in value_sets:
            for system in value_set.values:
                if system not in codes:
                    codes[system] = {'system': system, 'code': list()}

                codes[system]['code'].extend(list(value_set.values[system]))

        return {
            'type': command_type,
            'filter': {
                'coding': [val for key, val in codes.items()],
            },
        }


class AllergyRecommendation(Recommendation):
    """
    A recommendation specifying an allergy to add to the patient's allergy list.
    """

    def __init__(
            self, key, allergy: Type[ValueSet], rank=RANK_MAX, button='Allergy',
            title=None, narrative=None, context=None):
        if not title:
            title = allergy.name
        if not narrative:
            narrative = ''

        command = self.command_filter(COMMAND_TYPE_ALLERGY, [allergy])

        super().__init__(
            key, rank, button, title=title, narrative=narrative, command=command, context=context
        )


class DiagnoseRecommendation(Recommendation):
    """
    A recommendation specifying a condition to diagnose the patient.
    """

    def __init__(
            self, key, patient: Patient, condition: Type[ValueSet],
            rank=RANK_MAX, button='Diagnose', title=None, narrative=None, context=None):
        if not title:
            title = condition.name
        if not narrative:
            narrative = f'Consider diagnosing {patient.first_name} with {condition.name}'

        command = self.command_filter(COMMAND_TYPE_DIAGNOSE, [condition])

        super().__init__(
            key, rank, button, title=title, narrative=narrative, command=command, context=context)


class FollowUpRecommendation(Recommendation):
    """
    A recommendation specifying a follow-up appointment for the patient.
    """

    def __init__(
            self, key, patient: Patient, rank=RANK_MAX, button='Follow up', title=None,
            narrative=None, context=None):
        if not title:
            title = 'Request follow-up appointment'
        if not narrative:
            narrative = f'{patient.first_name} needs a follow-up appointment'

        command = {'type': COMMAND_TYPE_FOLLOWUP}

        super().__init__(key, rank, button, title, narrative=narrative, command=command, context=context)


class ImagingRecommendation(Recommendation):
    """
    A recommendation specifying an imaging to order. Title and narrative are
    auto-generated but may be overridden.
    """

    def __init__(self,
                 key,
                 patient: Patient,
                 imaging,
                 rank=RANK_MAX,
                 button='Order',
                 title=None,
                 narrative=None,
                 context=None):
        if not title:
            title = f'Order {imaging.name}'
            narrative = f'{patient.first_name} should be ordered {imaging.name}.'

        command = self.command_filter(COMMAND_TYPE_IMAGING, [imaging])

        super().__init__(
            key, rank, button, title=title, narrative=narrative, command=command, context=context)


class ImmunizationRecommendation(Recommendation):
    """
    A recommendation specifying an immunization to order. Use this to guide the user to the
    [immunization command](https://canvas-medical.zendesk.com/hc/en-us/articles/360057140293-Documenting-an-Immunization)
    """

    def __init__(self,
                 key,
                 patient: Patient,
                 immunization: Type[ValueSet],
                 rank=RANK_MAX,
                 button='Order',
                 title=None,
                 narrative=None,
                 context=None):
        # use patient, condition and lab to generate text and commands
        if not title:
            title = f'Order {immunization.name}'
            narrative = f'{patient.first_name} should get an {immunization.name}.'

        command = self.command_filter(COMMAND_TYPE_IMMUNIZATION, [immunization])

        super().__init__(key, rank, button, title=title, narrative=narrative, command=command, context=context)


class InstructionRecommendation(Recommendation):
    """
    A recommendation specifying an instruction for the patient.
    """

    def __init__(self,
                 key,
                 patient: Patient,
                 instruction: Type[ValueSet],
                 rank=RANK_MAX,
                 button='Instruct',
                 title=None,
                 narrative=None,
                 context=None):
        # use patient, condition and lab to generate text and commands
        if not title:
            title = f'Instruct {instruction.name}'
            narrative = f'{patient.first_name} should get an {instruction.name}.'

        command = self.command_filter(COMMAND_TYPE_INSTRUCTION, [instruction])

        super().__init__(key, rank, button, title=title, narrative=narrative, command=command, context=context)


class InterviewRecommendation(Recommendation):
    """
    A recommendation specifying an interview for the patient.
    """

    # questionnaire should be a set of ValueSet
    def __init__(self,
                 key,
                 patient: Patient,
                 questionnaires: List[Type[ValueSet]],
                 rank=RANK_MAX,
                 button='Plan',
                 title=None,
                 narrative=None,
                 context=None):
        # use patient, condition and lab to generate text and commands
        if not title:
            labels = ', '.join([item.name for item in questionnaires])
            title = 'Interview %s' % labels
            narrative = '%s should be given a %s.' % (patient.first_name, labels)

        command = self.command_filter(COMMAND_TYPE_INTERVIEW, questionnaires)

        super().__init__(key, rank, button, title=title, narrative=narrative, command=command, context=context)


class LabRecommendation(Recommendation):
    """
    A recommendation specifying a lab to order. Title and narrative are auto-generated but
    may be overridden.
    """

    def __init__(self,
                 key: str,
                 patient: Patient,
                 condition: Type[ValueSet],
                 lab: Type[ValueSet],
                 rank: int = RANK_MAX,
                 button: str = 'Order',
                 title=None,
                 narrative=None,
                 context=None):
        # use patient, condition and lab to generate text and commands
        if not title:
            title = f'Order {lab.name}'
            narrative = (
                f'{patient.first_name} has {condition.name} and a {lab.name} is recommended.')

        command = self.command_filter(COMMAND_TYPE_LAB_ORDER, [lab])

        super().__init__(
            key, rank, button, title=title, narrative=narrative, command=command, context=context)


class PerformRecommendation(Recommendation):
    """
    A recommendation specifying a procedure to perform. Title and narrative are
    auto-generated but may be overridden.
    """

    def __init__(self,
                 key: str,
                 patient: Patient,
                 procedure: Type[ValueSet],
                 condition: Type[ValueSet],
                 rank: int = RANK_MAX,
                 button: str = 'Perform',
                 title: Optional[str] = None,
                 narrative: Optional[str] = None,
                 context: Optional[Dict] = None):
        if not title:
            title = f'Perform {procedure.name}'
            narrative = (f'{patient.first_name} has {condition.name} and a {procedure.name} is '
                         'recommended.')

        command = self.command_filter(COMMAND_TYPE_PERFORM, [procedure])

        super().__init__(key, rank, button, title=title, narrative=narrative, command=command, context=context)


class PlanRecommendation(Recommendation):
    """
    A recommendation advising the provider to make a plan for the patient.
    """

    def __init__(
            self, key, patient: Patient, rank=RANK_MAX, button='Plan', title=None, narrative=None,
            context=None):
        # use patient, condition and lab to generate text and commands
        if not title:
            title = 'Make a plan'
            narrative = f'{patient.first_name} should have a plan.'

        command = {'type': COMMAND_TYPE_PLAN}

        super().__init__(key, rank, button, title=title, narrative=narrative, command=command, context=context)


class PrescribeRecommendation(Recommendation):
    """
    A recommendation specifying an prescribe for the patient.
    """

    def __init__(self,
                 key,
                 patient: Patient,
                 prescription: Type[ValueSet],
                 rank=RANK_MAX,
                 button: str = 'Prescribe',
                 title=None,
                 narrative=None,
                 context=None):
        # use patient, condition and lab to generate text and commands
        if not title:
            title = f'Prescribe {prescription.name}'
            narrative = f'{patient.first_name} should be prescribed {prescription.name}.'

        command = self.command_filter(COMMAND_TYPE_PRESCRIBE, [prescription])

        super().__init__(key, rank, button, title=title, narrative=narrative, command=command, context=context)


class ReferRecommendation(Recommendation):
    """
    A recommendation specifying a referral to order. Title and narrative are
    auto-generated but may be overridden.
    """

    def __init__(self,
                 key: str,
                 patient: Patient,
                 referral: Type[ValueSet],
                 condition: Optional[Type[ValueSet]] = None,
                 rank: int = RANK_MAX,
                 button: str = 'Refer',
                 title=None,
                 narrative=None,
                 context=None):
        if not title:
            title = f'Refer {referral.name}'

            if condition:
                narrative = (
                    f'{patient.first_name} has {condition.name} and should be referred for '
                    f'{referral.name}.')
            else:
                narrative = f'{patient.first_name} should be referred for {referral.name}.'

        command = self.command_filter(COMMAND_TYPE_REFERRAL, [referral])

        super().__init__(
            key, rank, button, title=title, narrative=narrative, command=command, context=context)


class StructuredAssessmentRecommendation(Recommendation):
    """
    StructuredAssessmentRecommendation

    A recommendation specifying a structured assessment for the patient.
    """

    # questionnaire should be a set of ValueSet
    def __init__(self,
                 key,
                 patient: Patient,
                 questionnaires: List[Type[ValueSet]],
                 rank=RANK_MAX,
                 button: str = 'Assess',
                 title=None,
                 narrative=None,
                 context=None):
        if not title:
            labels = ', '.join([item.name for item in questionnaires])
            title = f'Interview {labels}'
            narrative = f'{patient.first_name} should be given a {labels}.'

        command = self.command_filter(COMMAND_TYPE_STRUCTURED_ASSESSMENT, questionnaires)

        super().__init__(key, rank, button, title=title, narrative=narrative, command=command, context=context)


class TaskRecommendation(Recommendation):
    """
    A recommendation advising the provider to setup a task for someone else.
    """

    def __init__(self, patient, *args, **kwargs):
        super().__init__(command={'type': COMMAND_TYPE_TASK}, *args, **kwargs)


class VitalSignRecommendation(Recommendation):
    """
    A recommendation specifying a vital sign reading should be taken.
    """

    def __init__(
            self, key, patient: Patient, rank=RANK_MAX, button='Plan', title=None, narrative=None,
            context=None):
        # use patient, condition and lab to generate text and commands
        if not title:
            title = 'Collect vitals'
            narrative = f'{patient.first_name} should vital signs collected.'

        command = {'type': COMMAND_TYPE_VITALSIGN}

        super().__init__(key, rank, button, title=title, narrative=narrative, command=command, context=context)


class HyperlinkRecommendation(Recommendation):
    """
    A recommendation specifying a button and href that opens a link in a new tab.
    """

    def __init__(self, key, href, title='', button='', rank=RANK_MAX):
        super().__init__(key, rank, button, title=title, href=href)



import json
import os

from unittest import TestCase

from pathlib import Path

from canvas_workflow_kit.patient import Patient
from canvas_workflow_kit.protocol import ProtocolResult
from canvas_workflow_kit.utils import load_local_patient


class SDKBaseTest(TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            from canvas_workflow_kit import settings
            self.mocks_path = f'{settings.BASE_DIR}/tests/mock_data'
        except:
            self.mocks_path = '.'

    def load_patient(self, patient_key) -> Patient:

        patient_path = Path(f'{self.mocks_path}/{patient_key}/')

        assert patient_path.is_dir(), f'no patient directory exists from {patient_path}'

        patient = load_local_patient(patient_path)
        patient.patient['key'] = patient_key
        return patient

    def load_patient_data(self, patient_key, field):
        """
        Load data from mock data JSON files dumped by the dump_patient command.
        """
        filename = f'{self.mocks_path}/{patient_key}/{field}.json'

        if not os.path.exists(filename):
            if field == 'patient':
                raise Exception(f'Missing mock patient data for {patient_key}!')

            return []

        with open(filename, 'r') as fh:
            return json.load(fh)

    def assertIsNotApplicable(self, result: ProtocolResult):
        self.assertEqual('not_applicable', result.status)
        self.assertEqual('', result.narrative)
        self.assertEqual([], result.recommendations)
        self.assertIsNone(result.due_in)



from canvas_workflow_kit.protocol import ClinicalQualityMeasure, ProtocolResult
from canvas_workflow_kit.recommendation import Recommendation


class ProtoUnitTestAllTrue(ClinicalQualityMeasure):

    class Meta:

        references = [
            'This is the proto test all true. www.unittests.med/protounittest?id=alltrue',
        ]
        responds_to_event_types = [
            'HEALTH_MAINTENANCE',
            'UNIT_TESTS',
        ]
        identifiers = ['ProtoUnitTestAllTrue']

        compute_on_change_types = []

    def in_initial_population(self) -> bool:
        return True

    def in_denominator(self) -> bool:
        return True

    def in_numerator(self) -> bool:
        return True

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        result.status = 'Current'
        result.add_narrative('This is the narrative for the proto test all true')
        result.add_recommendation(TestAResult())
        result.add_recommendation(TestBResult())
        result.due_in = 17
        return result


class ProtoUnitTestDisabled(ClinicalQualityMeasure):

    class Meta:

        references = [
            'This is the proto test X. www.unittests.med/protounittest?id=X',
        ]
        responds_to_event_types = [
            'HEALTH_MAINTENANCE',
            'UNIT_TESTS',
        ]
        identifiers = ['ProtoUnitTestDisabled']

    @classmethod
    def enabled(cls) -> bool:
        return False

    def in_initial_population(self) -> bool:
        return True

    def in_denominator(self) -> bool:
        return True

    def in_numerator(self) -> bool:
        return True

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        result.status = 'Current'
        result.add_narrative('This is the narrative for the proto test x')
        result.add_recommendation(TestAResult())
        result.add_recommendation(TestBResult())
        result.due_in = 5
        return result


class ProtoUnitTestX(ClinicalQualityMeasure):
    class Meta:

        references = [
            'This is the proto test X. www.unittests.med/protounittest?id=X',
        ]
        responds_to_event_types = [
            'UNIT_TESTS',
        ]
        identifiers = ['ProtoUnitTestX']

        compute_on_change_types = [
            'change_C',
        ]

    def in_initial_population(self) -> bool:
        return True

    def in_denominator(self) -> bool:
        return True

    def in_numerator(self) -> bool:
        return True

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        result.status = 'Current'
        result.add_narrative('This is the narrative for the proto test x')
        result.add_recommendation(TestBResult())
        result.due_in = -3
        return result


class ProtoUnitTestY(ClinicalQualityMeasure):

    class Meta:

        references = [
            'This is the proto test Y. www.unittests.med/protounittest?id=Y',
        ]
        responds_to_event_types = [
            'HEALTH_MAINTENANCE',
            'UNIT_TESTS',
        ]
        identifiers = ['ProtoUnitTestY']

    def in_initial_population(self) -> bool:
        return True

    def in_denominator(self) -> bool:
        return True

    def in_numerator(self) -> bool:
        return True

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        result.status = 'not_applicable'
        result.add_narrative('This is the narrative for the proto test Y')
        result.add_recommendation(TestAResult())
        result.due_in = None
        return result


class ProtoUnitTestZ(ClinicalQualityMeasure):
    class Meta:

        references = [
            'This is the proto test Z. www.unittests.med/protounittest?id=Y',
        ]
        responds_to_event_types = [
            'HEALTH_MAINTENANCE',
            'UNIT_TESTS',
        ]
        identifiers = ['ProtoUnitTestZ']

        compute_on_change_types = [
            'change_B',
        ]

    def in_initial_population(self) -> bool:
        return True

    def in_denominator(self) -> bool:
        return True

    def in_numerator(self) -> bool:
        return True

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        result.status = 'not_applicable'
        result.add_narrative('This is the narrative for the proto test Y')
        result.add_recommendation(TestAResult())
        result.due_in = 33
        return result


class TestAResult(Recommendation):

    def __init__(self):
        command = {'type': 'TestAResult'}

        super().__init__(
            key='KEY-AR',
            rank=123,
            button='ACT',
            title='Title TestAResult',
            narrative='Narrative TestAResult',
            command=command)


class TestBResult(Recommendation):

    def __init__(self):
        command = {'type': 'TestBResult'}

        super().__init__(
            key='KEY-BR',
            rank=456,
            button='RUN',
            title='Title TestBResult',
            narrative='Narrative TestBResult',
            command=command)
